"Compatibility layer for various Python versions"
from __future__ import annotations

import sys
from itertools import tee
from typing import TYPE_CHECKING, Callable, Iterable, Iterator, TypeVar

__all__ = ["pairwise", "cache"]


if sys.version_info < (3, 10) or TYPE_CHECKING:  # pragma: no cover
    T = TypeVar("T")

    def pairwise(i: Iterable[T]) -> Iterator[tuple[T, T]]:
        a, b = tee(i)
        next(b, None)
        return zip(a, b)

else:
    from itertools import pairwise


if sys.version_info < (3, 9) or TYPE_CHECKING:  # pragma: no cover
    from functools import lru_cache

    _Tcall = TypeVar("_Tcall", bound=Callable[..., object])

    def cache(func: _Tcall) -> _Tcall:
        return lru_cache(maxsize=None)(func)  # type: ignore

else:
    from functools import cache



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from itertools import tee
from typing import (
    ClassVar,
    Iterable,
    Iterator,
    Literal,
    Protocol,
    Sequence,
    cast,
    final,
)

from ..common import XY, Align, Pt, add_slots, advance, prepend, setattr_frozen
from ..resources import Resources
from ..style import Span, Style, StyledMixin, StyleFull, StyleLike
from ..typeset import firstfit, optimum
from ..typeset.layout import ShapedText
from ..typeset.parse import into_words
from ..typeset.state import Passage, State, max_lead, splitlines
from ..typeset.words import WordLike, indent_first
from .common import Block, ColumnFill


@add_slots
@dataclass(frozen=True)
class LinebreakParams:
    """Parameters for tweaking the optimum-fit algorithm.

    Parameters
    ----------
    tolerance
        The tolerance for the stretch of each line.
        If no feasible solution is found, the tolerance is increased until
        there is.
        Increase the tolerance if you want to avoid hyphenation
        at the cost of more stretching and longer runtime.
    hyphen_penalty
        The penalty for hyphenating a word. If increasing this value does
        not result in fewer hyphens, try increasing the tolerance.
    consecutive_hyphen_penalty
        The penalty for placing hyphens on consecutive lines. If increasing
        this value does not appear to work, try increasing the tolerance.
    fitness_diff_penalty
        The penalty for very tight and very loose lines following each other.
    """

    tolerance: float = 1
    hyphen_penalty: float = 1000
    consecutive_hyphen_penalty: float = 1000
    fitness_diff_penalty: float = 1000

    DEFAULT: ClassVar["LinebreakParams"]


LinebreakParams.DEFAULT = LinebreakParams()


@final
@add_slots
@dataclass(frozen=True, init=False)
class Paragraph(Block, StyledMixin):
    """A :class:`Block` that renders a paragraph of text.

    Parameters
    ----------
    content
        The text to render. Can be a string, or a nested :class:`~pdfje.Span`.
    style
        The style to render the text with.
        See :ref:`tutorial<style>` for more details.
    align: Align | ``"left"`` | ``"center"`` | ``"right"`` | ``"justify"``
        The horizontal alignment of the text.
    indent
        The amount of space to indent the first line of the paragraph.
    avoid_orphans
        Whether to avoid orphans (single lines before or after a page or
        column break).
    optimal
        Whether to use the optimal paragraph layout algorithm.
        If set to ``False`` or ``None``, a faster but less optimal algorithm
        is used. To customize the algorithm parameters, pass an
        :class:`~pdfje.layout.LinebreakParams` object.

    Examples
    --------

    .. code-block:: python

        from pdfje.layout import Paragraph
        from pdfje.style import Style

        Paragraph(
            "This is a paragraph of text.",
            style="#003311",
            align="center",
        )
        Paragraph(
            [
                "It can also be ",
                Span("styled", "#00ff00"),
                " in multiple ",
                Span("ways", Style(size=14, italic=True)),
                ".",
            ],
            style=times_roman,
            optimal=False,
        )

    """

    content: Sequence[str | Span]
    style: Style
    align: Align
    indent: Pt
    avoid_orphans: bool
    optimal: LinebreakParams | None

    def __init__(
        self,
        content: str | Span | Sequence[str | Span],
        style: StyleLike = Style.EMPTY,
        align: Align
        | Literal["left", "center", "right", "justify"] = Align.LEFT,
        indent: Pt = 0,
        avoid_orphans: bool = True,
        optimal: LinebreakParams | bool | None = True,
    ):
        if isinstance(content, (str, Span)):
            content = [content]
        setattr_frozen(self, "content", content)
        setattr_frozen(self, "style", Style.parse(style))
        setattr_frozen(self, "align", Align.parse(align))
        setattr_frozen(self, "indent", indent)
        setattr_frozen(self, "avoid_orphans", avoid_orphans)
        if isinstance(optimal, bool):
            optimal = LinebreakParams.DEFAULT if optimal else None
        setattr_frozen(self, "optimal", optimal)

    def into_columns(
        self, res: Resources, style: StyleFull, cs: Iterator[ColumnFill]
    ) -> Iterator[ColumnFill]:
        style |= self.style
        state = style.as_state(res)
        passages = list(self.flatten(res, style))
        lead = max_lead(passages, state)
        col = next(cs)
        for para in splitlines(passages):
            cs, _branch = tee(prepend(col, cs))
            [*filled, col], state = _fill_paragraph(
                iter(para),
                _branch,
                state,
                self.indent,
                lead,
                self.align,
                self.avoid_orphans,
                shape=cast(
                    Shaper,
                    (partial(optimum.shape, params=self.optimal))
                    if self.optimal
                    else firstfit.shape,
                ),
            )
            advance(cs, len(filled) + 1)
            yield from filled
        yield col


class Shaper(Protocol):
    def __call__(
        self,
        ws: Iterator[WordLike],
        columns: Iterator[XY],
        allow_empty: bool,
        lead: Pt,
        avoid_orphans: bool,
        align: Align,
    ) -> Iterator[ShapedText]:
        ...


def _fill_paragraph(
    txt: Iterator[Passage],
    cs: Iterator[ColumnFill],
    state: State,
    indent: Pt,
    lead: Pt,
    align: Align,
    avoid_orphans: bool,
    shape: Shaper,
) -> tuple[Iterable[ColumnFill], State]:
    done: list[ColumnFill] = []
    col = next(cs)
    allow_empty = bool(col.blocks)
    cmd, words = into_words(txt, state)
    state = cmd.apply(state)
    cs, _branch = tee(prepend(col, cs))
    for chunk, col in zip(  # pragma: no branch
        shape(
            indent_first(words, indent),
            (XY(c.box.width, c.height_free) for c in _branch),
            allow_empty=allow_empty,
            lead=lead,
            align=align,
            avoid_orphans=avoid_orphans,
        ),
        cs,
    ):
        done.append(col.add(chunk))
        state = chunk.end_state() or state
    return done, state



from __future__ import annotations

from dataclasses import dataclass, field
from itertools import chain, count
from typing import Iterable, Iterator

from . import atoms
from .atoms import ASCII
from .common import add_slots
from .fonts import BuiltinTypeface, TrueType
from .fonts.common import BuiltinFont, Font, Typeface
from .fonts.embed import OBJS_PER_EMBEDDED_FONT, Subset


@add_slots
@dataclass(frozen=True, eq=False)
class Resources:
    """Keeps track of PDF resources within a document, such as fonts"""

    _builtins: dict[tuple[ASCII, bool, bool], BuiltinFont] = field(
        default_factory=dict
    )
    _subsets: dict[tuple[TrueType, bool, bool], Subset] = field(
        default_factory=dict
    )
    _next_subset_index: Iterator[int] = field(default_factory=count.__call__)

    def to_objects(self, first_id: atoms.ObjectID) -> Iterable[atoms.Object]:
        for sub, i in zip(
            self._subsets.values(),
            count(first_id, step=OBJS_PER_EMBEDDED_FONT),
        ):
            yield from sub.to_objects(i)

    def to_atoms(self, first_id: atoms.ObjectID) -> atoms.Dictionary:
        return atoms.Dictionary(
            (
                b"Font",
                atoms.Dictionary(
                    *chain(
                        (
                            (b.id, b.to_resource())
                            for b in self._builtins.values()
                        ),
                        (
                            (s.id, atoms.Ref(obj_id))
                            for s, obj_id in zip(
                                self._subsets.values(),
                                count(first_id, step=OBJS_PER_EMBEDDED_FONT),
                            )
                        ),
                    )
                ),
            )
        )

    def font(self, f: Typeface, bold: bool, italic: bool) -> Font:
        if isinstance(f, BuiltinTypeface):
            return self._builtins.setdefault(
                (f.regular.name, bold, italic), f.font(bold, italic)
            )
        else:
            try:
                return self._subsets[(f, bold, italic)]
            except KeyError:
                new_subset = self._subsets[(f, bold, italic)] = Subset.new(
                    b"F%i" % next(self._next_subset_index),
                    f.font(bold, italic),
                )
                return new_subset



from __future__ import annotations

from functools import partial
from operator import mul
from typing import Callable

from .common import XY, Pt

__all__ = [
    "inch",
    "cm",
    "mm",
    "pc",
    "pt",
    "Pt",
]

inch: Callable[[float], Pt] = partial(mul, 72)
inch.__doc__ = "Convert inches to points"
cm: Callable[[float], Pt] = partial(mul, 28.346456692913385)
cm.__doc__ = "Convert centimeters to points"
mm: Callable[[float], Pt] = partial(mul, 2.8346456692913385)
mm.__doc__ = "Convert millimeters to points"
pc: Callable[[float], Pt] = partial(mul, 12)
pc.__doc__ = "Convert picas to points"


def pt(x: float) -> Pt:
    "No-op conversion. Can be used to make units explicit."
    return x


A0 = XY(2380, 3368)
"A0 paper size"
A1 = XY(1684, 2380)
"A1 paper size"
A2 = XY(1190, 1684)
"A2 paper size"
A3 = XY(842, 1190)
"A3 paper size"
A4 = XY(595, 842)
"A4 paper size"
A5 = XY(420, 595)
"A5 paper size"
A6 = XY(297, 420)
"A6 paper size"
letter = XY(612, 792)
"Letter paper size"
legal = XY(612, 1008)
"Legal paper size"
tabloid = XY(792, 1224)
"Tabloid paper size"
ledger = tabloid.flip()
"Ledger paper size, same as tabloid landscape"



"A simple first-fit line wrapping algorithm."
from __future__ import annotations

from dataclasses import dataclass
from itertools import tee
from typing import Iterable, Iterator, NamedTuple, Sequence

from ..atoms import LiteralStr, Real
from ..common import XY, Align, NonEmptyIterator, Pt, add_slots, prepend
from .layout import Line as _Line
from .layout import ShapedText
from .words import WordLike, render_kerned


def shape(
    words: Iterator[WordLike],
    columns: Iterator[XY],
    allow_empty: bool,
    lead: Pt,
    avoid_orphans: bool,
    align: Align,
) -> Iterator[ShapedText]:
    _shape = _shape_avoid_orphans if avoid_orphans else _shape_simple
    return (
        (
            ShapedText(
                list(map(Line.justify, ls)), lead, align, len(ls) * lead
            )
            for ls in _shape(words, columns, allow_empty, lead)
        )
        if align is Align.JUSTIFY
        else (
            ShapedText(ls, lead, align, len(ls) * lead)
            for ls in _shape(words, columns, allow_empty, lead)
        )
    )


def _shape_avoid_orphans(
    ws: Iterator[WordLike] | None,
    columns: Iterator[XY],
    allow_empty: bool,
    lead: Pt,
) -> Iterator[Sequence[Line]]:
    col = next(columns)
    ws, lines, ws_undo = take_box(ws, col, allow_empty, lead)
    # In case of an avoidable orphan, start over
    if ws and len(lines) == 1 and allow_empty:
        ws = ws_undo
        lines = ()
    elif not ws:
        yield lines
        return

    col = next(columns)
    while True:
        lines_prev = lines
        ws_undo_prev = ws_undo
        ws, lines, ws_undo = take_box(ws, col, False, lead)
        # case: paragraph not done. Continue to next column.
        if ws:
            yield lines_prev
            col = next(columns)
        # case: a potentially fixable orphan
        elif len(lines) == 1 and len(lines_prev) > 2 and col.y >= lead * 2:
            # FUTURE: optimize the case where the column widths are the same,
            #         and we don't need to re-typeset the last line.
            assert ws_undo_prev is not None
            ws_undo_prev, _branch = tee(ws_undo_prev)
            _, _lines_new, ws_undo = take_box(_branch, col, False, lead)
            if len(_lines_new) == 1:
                break  # our attempt to fix the orphan failed. We're done.
            else:
                lines = lines_prev[:-1]
                ws = ws_undo_prev
        # case: we're done, but no (fixable) orphan.
        else:
            break

    yield lines_prev
    yield lines


# filling is a lot simpler if we don't avoid orphaned lines.
def _shape_simple(
    ws: Iterator[WordLike] | None,
    columns: Iterator[XY],
    allow_empty: bool,
    lead: Pt,
) -> Iterator[Sequence[Line]]:
    for col in columns:  # pragma: no branch
        ws, lines, _ = take_box(ws, col, allow_empty, lead)
        yield lines
        if not ws:
            return
        allow_empty = False


class _FilledBox(NamedTuple):
    rest: NonEmptyIterator[WordLike] | None
    lines: Sequence[Line]
    rest_incl_lastline: NonEmptyIterator[WordLike] | None


def take_box(
    queue: NonEmptyIterator[WordLike] | None,
    space: XY,
    allow_empty: bool,
    lead: Pt,
) -> _FilledBox:
    width, height = space
    max_line_count: float = height // lead or not allow_empty
    lines: list[Line] = []
    queue_prev = queue
    while queue and len(lines) < max_line_count:
        # OPTIMIZE: it'd be more efficient to only 'tee' on the last line
        queue, queue_prev = tee(queue)
        queue, ln = take_line(queue, width)
        lines.append(ln)
    return _FilledBox(queue, lines, queue_prev)


def take_line(
    ws: NonEmptyIterator[WordLike], width: Pt
) -> tuple[NonEmptyIterator[WordLike] | None, Line]:
    space = width
    content: list[WordLike] = []

    for word in ws:
        if word.pruned_width() > space:
            break

        space -= word.width
        content.append(word)
    else:
        # i.e. this is the last line of the paragraph
        return (None, Line(tuple(content), width - space, 0))

    last_word, dangling = word.hyphenate(space)
    ws = prepend(dangling, ws)
    if last_word:
        space -= last_word.width
        content.append(last_word)
    elif content and (extra_space := content[-1].prunable_space()):
        content[-1] = content[-1].pruned()
        space += extra_space
    elif not content:
        # We force placing at least a minimal word fragment to avoid
        # infinitely waiting for enough width.
        # This shouldn't occur in practice often, where the column
        # width is much larger than the longest word segment.
        word, leftover = next(ws).minimal_box()
        if leftover:
            ws = prepend(leftover, ws)
        else:
            # An extra check is needed to tell whether this is the last
            # word in the paragraph.
            try:
                ws = prepend(next(ws), ws)
            except StopIteration:
                return (None, Line((word,), word.width, 0))
        content = [word]
        space -= word.width

    return (ws, Line(tuple(content), width - space, space))


@add_slots
@dataclass(frozen=True)
class Line(_Line):
    words: tuple[WordLike, ...]
    width: Pt
    space: Pt

    def justify(self) -> Line:
        try:
            # The additional width per word break, weighted by the font size,
            # which is needed to justify the text.
            width_per_break = self.space / sum(
                w.state.size for w in self.words if w.tail
            )
        except ZeroDivisionError:
            return self  # No word breaks means no justification.
        return Line(
            tuple(
                w.extend_tail(width_per_break * w.state.size)
                for w in self.words
            ),
            self.width + self.space,
            0,
        )

    def __iter__(self) -> Iterator[bytes]:
        content: Iterable[Real | LiteralStr] = ()
        for w in self.words:
            content = yield from w.encode_into_line(content)
        yield from render_kerned(content)



"""Knuth-Plass line breaking algorithm.

Changes from the original algorithm:
- Instead of generic glue and penalty, we use explicit spaces and hyphenation.
- We don't allow for forced breaks within a paragraph -- only at the end.
"""
from __future__ import annotations

from dataclasses import dataclass, field
from operator import attrgetter
from typing import Callable, Iterable, Literal, Sequence

from ..common import add_slots, peek
from ..compat import pairwise

Pos = int  # position in the list of boxes
LineNum = int  # zero-based line number
# As recommended in the paper, we store most widths cumulatively
# to prevent summing over and over
CumulativeWidth = float
# The adjustment ratio to justify the line, where -1 is
# fully shrunk and 1 is fully stretched. Note that stretch > 1 is possible
# (if potentially ugly), but additional shrink is not
# because words would be too close together.
Ratio = float
BIG: Ratio = 100
# Tight, loose, very tight, very loose
Fitness = Literal[0, 1, 2, 3]


@add_slots
@dataclass(frozen=True)
class Box:
    measure: CumulativeWidth
    stretch: CumulativeWidth
    shrink: CumulativeWidth
    incl_space: CumulativeWidth

    incl_hyphen: CumulativeWidth
    hyphenated: bool
    no_break: bool


def optimum_fit(
    bs: Iterable[Box],
    width: Callable[[LineNum], float],
    tol: float,
    hyphen_penalty: float = 100,
    consecutive_hyphen_penalty: float = 3000,
    fit_diff_penalty: float = 100,
    ragged: bool = False,
) -> Sequence[Break]:
    """Return the optimal break points for justifying the given boxes.

    Optimal means:
    - Minimizing the total stretch/shrink of lines
    - Avoiding hyphens if possible
    - Avoiding consecutive hyphens if possible
    - Avoiding very tight and very loose lines from following each other

    Parameters
    ----------
    bs
        The boxes to break.
    width
        A function that returns the width of the line at the given line number.
        Line numbers start at 0.
    tol
        The tolerance for the adjustment ratio. If no feasible breaks are found
        for the given tolerance, a `NoFeasibleBreaks` exception is raised.
        A higher tolerance will still produce an optimal result,
        but it will take longer to compute and may contain
        more 'ugly' lines (if this causes the overall result to be better).
    hyphen_penalty
        The penalty for a hyphen at the end of a line.
    consecutive_hyphen_penalty
        The penalty for consecutive hyphens.
    fit_diff_penalty
        The penalty for a difference in fitness between two consecutive breaks.
        (for example a very loose line followed by a tight line)
    ragged
        The algorithm is slightly different depending on whether you need
        ragged or justified text.
        If `ragged` is `False` (i.e. justtified), lines are penalized
        for stretching/shrinking of space *between* words.
        If `ragged` is `True`, lines are penalized for unused space,
        regardless of how many spaces are available between words.

        Note
        ----

        If ragged is set to `True`, the `stretch` value of the boxes
        should all be the same, and the `shrink` value should be 0.
        A good value for `stretch` is 3 times the space width.
        (This is derived from the approach in the paper)
    """
    # FUTURE: The ragged case probably be optimized further -- by eliminating
    #         the fitness difference penalty, for example.
    ratio = ratio_ragged if ragged else ratio_justified
    g = _BreakNetwork()

    try:
        box_next, bs = peek(iter(bs))
    except StopIteration:
        return ()

    pos = 0
    for pos, (box, box_next) in enumerate(pairwise(bs), start=1):
        if box.no_break:
            continue
        if g.is_empty():
            # Breaking here saves us time needlessly looping over the boxes
            # once we know there are no feasible breaks
            raise NoFeasibleBreaks()

        for node in list(g.nodes()):
            r = ratio(
                node.measure, node.stretch, node.shrink, box, width(node.line)
            )
            if r < -1:
                # This break is no longer feasible, because the line
                # would have to shrink too much to accomodate the content.
                # Thus, we remove it.
                g.remove(node)
            elif r <= tol:
                fit = _fitness(r)
                g.add(
                    _BreakNode(
                        pos,
                        node.line + 1,
                        r,
                        fit,
                        (
                            _main_demerit(hyphen_penalty * box.hyphenated, r)
                            + (
                                consecutive_hyphen_penalty
                                * box.hyphenated
                                * node.hyphenated
                            )
                            + (abs(fit - node.fitness) > 1) * fit_diff_penalty
                            + node.demerits
                        ),
                        box.incl_space,
                        box_next.stretch,
                        box_next.shrink,
                        box.hyphenated,
                        node,
                    )
                )

    return _optimal_end(
        g.nodes(), box_next, pos + 1, width, fit_diff_penalty, ratio
    ).unroll()


@add_slots
@dataclass(frozen=True)
class Break:
    pos: Pos
    adjust: Ratio
    demerits: float
    width: float


class NoFeasibleBreaks(Exception):
    "Raised when there are no feasible breaks for the given tolerance."


def _optimal_end(
    nodes: Iterable[_BreakNode],
    box: Box,
    pos: int,
    width: Callable[[LineNum], float],
    fit_diff_penalty: float,
    ratio: Callable[
        [CumulativeWidth, CumulativeWidth, CumulativeWidth, Box, float], Ratio
    ],
) -> _EndNode:
    options = (
        _EndNode(
            pos,
            r,
            _main_demerit(0, r)
            + (abs(_fitness(r) - n.fitness) > 1) * fit_diff_penalty
            + n.demerits,
            box.incl_space,
            n,
        )
        for n in nodes
        if (
            r := min(
                ratio(n.measure, n.stretch, n.shrink, box, width(n.line)), 0
            )
        )
        > -1
        # We don't need to check for r <= tol,
        # because the last line doesn't need to stretch
    )

    try:
        return min(options, key=_demerits)
    except ValueError:  # i.e. options is empty
        raise NoFeasibleBreaks()


def _fitness(r: Ratio) -> Fitness:
    return 0 if r < -0.5 else 1 if r <= 0.5 else 2 if r <= 1 else 3


_demerits: Callable[[_EndNode], float] = attrgetter("demerits")


def ratio_justified(
    measure: CumulativeWidth,
    stretch: CumulativeWidth,
    shrink: CumulativeWidth,
    b: Box,
    space: float,
) -> Ratio:
    length = b.incl_hyphen - measure
    try:
        return (space - length) / (
            (b.stretch - stretch) if length < space else (b.shrink - shrink)
        )
    except ZeroDivisionError:
        return (BIG + length / space) if length != space else 0


def ratio_ragged(
    measure: CumulativeWidth,
    _: CumulativeWidth,
    __: CumulativeWidth,
    b: Box,
    space: float,
) -> Ratio:
    leftover = space - b.incl_hyphen + measure
    if leftover >= 0:
        return leftover / b.stretch
    else:
        # If there is not enough space, returning anything smaller than -1
        # will cause the break to be discarded.
        return -2


def _main_demerit(penalty: float, r: Ratio) -> float:
    return (1 + 100 * abs(r) ** 3 + penalty) ** 2


@add_slots
@dataclass(frozen=True)
class _EndNode:
    pos: Pos
    ratio: Ratio
    demerits: float
    measure: float
    prev: _BreakNode = field(repr=False)

    def unroll(self) -> Sequence[Break]:
        node: _BreakNode | _EndNode = self
        result: list[Break] = []
        while node is not _ROOT:
            result.append(
                Break(
                    node.pos,
                    node.ratio,
                    node.demerits - node.prev.demerits,
                    node.measure - node.prev.measure,
                )
            )
            node = node.prev
        result.reverse()
        return result


@add_slots
@dataclass(frozen=True)
class _BreakNode:
    pos: Pos
    line: LineNum
    ratio: Ratio
    fitness: Fitness
    demerits: float

    measure: CumulativeWidth
    stretch: CumulativeWidth
    shrink: CumulativeWidth

    hyphenated: bool

    prev: _BreakNode = field(repr=False)


_ROOT = _BreakNode(0, 0, 0, 1, 0, 0, 0, 0, False, NotImplemented)


class _BreakNetwork:
    "A directed acyclic graph of possible breaks."
    __slots__ = ("_inner",)

    def __init__(self) -> None:
        self._inner: dict[tuple[LineNum, Pos, Fitness], _BreakNode] = {
            (0, 0, 1): _ROOT
        }

    # FUTURE: does the sorting of the output matter?
    def nodes(self) -> Iterable[_BreakNode]:
        return self._inner.values()

    def add(self, n: _BreakNode) -> None:
        key = (n.line, n.pos, n.fitness)
        # OPTIMIZE: from the paper: "we need not remember the Class 0
        # possibility if its total demerits exceed those of the Class 2 break
        # plus the demerits for contrasting lines, since the Class 0
        # breakpoint will never be optimum in such a case."
        old = self._inner.get(key)
        if old is None or n.demerits < old.demerits:
            self._inner[key] = n

    def is_empty(self) -> bool:
        return not self._inner

    def remove(self, n: _BreakNode) -> None:
        self._inner.pop((n.line, n.pos, n.fitness))



from __future__ import annotations

import enum
from dataclasses import dataclass, fields
from itertools import chain, islice, tee
from operator import itemgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    final,
    overload,
)

Pt = float  # not allowed to be inf or nan (math.isfinite)
Char = str  # 1-character string
Pos = int  # position within a string (index)
HexColor = str  # 6-digit hex color, starting with `#`. e.g. #ff0000
Streamable = Iterable[bytes]  # PDF stream content

flatten = chain.from_iterable
first = itemgetter(0)
second = itemgetter(1)
Ordinal = int  # a unicode code point
NonEmptySequence = Sequence

Tclass = TypeVar("Tclass", bound=type)
T = TypeVar("T")
U = TypeVar("U")
NonEmptyIterator = Iterator[T]


def prepend(i: T, it: Iterable[T]) -> Iterator[T]:
    return chain((i,), it)


# From the itertools recipes
def advance(it: Iterator[object], n: int) -> None:
    "Advance the iterator n-steps ahead"
    next(islice(it, n, n), None)


# Abstract properties don't mix well with dataclass inheritance at runtime.
# Deleting properties at runtime fixes the issue.
# We still use the type checker to ensure subclasses have the right methods.
def fix_abstract_properties(t: Tclass) -> Tclass:
    for n in getattr(t, "__abstractmethods__", ()):
        try:
            delattr(t, n)
        except AttributeError:
            pass
    return t


@dataclass(frozen=True, repr=False)
class always(Generic[T]):
    __slots__ = ("_v",)

    _v: T

    def __call__(self, *_: Any, **__: Any) -> T:
        return self._v

    def __repr__(self) -> str:
        return f"always({self._v!r})"


def peek(it: Iterator[T]) -> tuple[T, Iterator[T]]:
    branch, it = tee(it)
    return next(branch), it


setattr_frozen = object.__setattr__


# adapted from github.com/ericvsmith/dataclasses
# under its Apache 2.0 license.
def add_slots(cls: Tclass) -> Tclass:  # pragma: no cover
    if "__slots__" in cls.__dict__:
        raise TypeError(f"{cls.__name__} already specifies __slots__")
    cls_dict = dict(cls.__dict__)
    field_names = tuple(f.name for f in fields(cls))
    cls_dict["__slots__"] = field_names
    for field_name in field_names:
        # Remove our attributes, if present. They'll still be
        #  available in _MARKER.
        cls_dict.pop(field_name, None)
    # Remove __dict__ itself.
    cls_dict.pop("__dict__", None)
    # And finally create the class.
    qualname = getattr(cls, "__qualname__", None)
    cls = type(cls)(cls.__name__, cls.__bases__, cls_dict)
    if qualname is not None:
        cls.__qualname__ = qualname
    return cls


@final
@add_slots
@dataclass(frozen=True, repr=False)
class XY(Sequence[float]):
    """Represents a point, vector, or size in 2D space, where the first
    coordinate is the horizontal component and the second is the vertical one.

    This class supports some basic operators:

    .. code-block:: python

        >>> XY(1, 2) + XY(3, 4)
        XY(4, 6)
        # two-tuples are also supported
        >>> XY(1, 2) - (3, 4)
        XY(-2, -2)
        >>> XY(1, 2) / 2
        XY(0.5, 1.0)
        >>> XY(1, 2) * 2
        XY(2, 4)

    It also implements the :class:`~collections.abc.Sequence` protocol:

    .. code-block:: python

        >>> xy = XY(1, 2)
        >>> xy[0]
        1
        >>> x, y = xy
        >>> list(xy)
        [1, 2]

    """

    x: float = 0
    y: float = 0

    def __repr__(self) -> str:
        return f"XY({self.x}, {self.y})"

    def astuple(self) -> tuple[float, float]:
        return (self.x, self.y)

    @staticmethod
    def parse(v: XY | tuple[float, float], /) -> XY:
        return XY(*v) if isinstance(v, tuple) else v

    def __iter__(self) -> Iterator[float]:
        yield self.x
        yield self.y

    # We don't support slices -- which is technically a Sequence protocol
    # violation. But in practice this is not an issue.
    def __getitem__(self, i: int) -> float:  # type: ignore[override]
        if i == 0:
            return self.x
        elif i == 1:
            return self.y
        else:
            raise IndexError(i)

    def __len__(self) -> int:
        return 2

    def __truediv__(self, other: float | int) -> XY:
        if isinstance(other, (float, int)):
            return XY(self.x / other, self.y / other)
        else:
            return NotImplemented  # type: ignore[unreachable]

    def __mul__(self, other: float | int) -> XY:
        if isinstance(other, (float, int)):
            return XY(self.x * other, self.y * other)
        else:
            return NotImplemented  # type: ignore[unreachable]

    def __sub__(self, other: XY | tuple[float, float]) -> XY:
        if isinstance(other, tuple):
            return XY(self.x - other[0], self.y - other[1])
        elif isinstance(other, XY):
            return XY(self.x - other.x, self.y - other.y)
        else:
            return NotImplemented  # type: ignore[unreachable]

    def __add__(self, other: XY | tuple[float, float]) -> XY:
        if isinstance(other, tuple):
            return XY(self.x + other[0], self.y + other[1])
        elif isinstance(other, XY):
            return XY(self.x + other.x, self.y + other.y)
        else:
            return NotImplemented  # type: ignore[unreachable]

    def add_x(self, x: float) -> XY:
        return XY(self.x + x, self.y)

    def add_y(self, y: float) -> XY:
        return XY(self.x, self.y + y)

    def flip(self) -> XY:
        "Return a new XY with x and y swapped"
        return XY(self.y, self.x)


class Align(enum.Enum):
    """Horizontal alignment of text."""

    LEFT = 0
    CENTER = 1
    RIGHT = 2
    JUSTIFY = 3

    @staticmethod
    def parse(align: Align | str) -> Align:
        if isinstance(align, str):
            return Align[align.upper()]
        return align


@add_slots
@dataclass(frozen=True)
class Sides(Sequence[float]):
    """Represents a set of four sides. Used for padding and margins."""

    top: Pt = 0
    right: Pt = 0
    bottom: Pt = 0
    left: Pt = 0

    def __iter__(self) -> Iterator[Pt]:
        yield self.top
        yield self.right
        yield self.bottom
        yield self.left

    def astuple(self) -> tuple[Pt, Pt, Pt, Pt]:
        return (self.top, self.right, self.bottom, self.left)

    # We don't support slices -- which is technically a Sequence protocol
    # violation. But in practice this is not an issue.
    def __getitem__(self, i: int) -> Pt:  # type: ignore[override]
        if i == 0:
            return self.top
        elif i == 1:
            return self.right
        elif i == 2:
            return self.bottom
        elif i == 3:
            return self.left
        else:
            raise IndexError(i)

    def __len__(self) -> int:
        return 4

    @staticmethod
    def parse(v: SidesLike, /) -> Sides:
        if isinstance(v, Sides):
            return v
        elif isinstance(v, tuple):
            if len(v) == 4:
                return Sides(*v)
            elif len(v) == 3:
                return Sides(v[0], v[1], v[2], v[1])  # type: ignore[misc]
            elif len(v) == 2:
                return Sides(v[0], v[1], v[0], v[1])
            else:
                raise TypeError(f"Cannot parse {v} as sides")
        elif isinstance(v, (float, int)):
            return Sides(v, v, v, v)
        else:
            raise TypeError(f"Cannot parse {v} as sides")


SidesLike = Union[
    Sides, Tuple[Pt, Pt, Pt, Pt], Tuple[Pt, Pt, Pt], Tuple[Pt, Pt], Pt
]


@final
@add_slots
@dataclass(frozen=True, repr=False)
class RGB(Sequence[float]):
    """Represents a color in RGB space, with values between 0 and 1.

    Common colors are available as constants:

    .. code-block:: python

        from pdfje import red, lime, blue, black, white, yellow, cyan, magenta

    Note
    ----

    In most cases where you would use a color, you can use a string
    of the form ``#RRGGBB`` or instead, e.g. ``#fa9225``

    """

    red: float = 0
    green: float = 0
    blue: float = 0

    def __post_init__(self) -> None:
        assert (
            self.red <= 1 and self.green <= 1 and self.blue <= 1
        ), "RGB values too large. They should be between 0 and 1"

    # We don't support slices -- which is technically a Sequence protocol
    # violation. But in practice this is not an issue.
    def __getitem__(self, i: int) -> float:  # type: ignore[override]
        if i == 0:
            return self.red
        elif i == 1:
            return self.green
        elif i == 2:
            return self.blue
        else:
            raise IndexError(i)

    def __len__(self) -> int:
        return 3

    @staticmethod
    def parse(v: RGB | tuple[float, float, float] | str, /) -> RGB:
        if isinstance(v, tuple):
            assert len(v) == 3, "RGB tuple must have 3 values"
            return RGB(*v)
        elif isinstance(v, str):
            assert v.startswith("#"), "RGB string must start with #"
            assert len(v) == 7, "RGB string must have 7 characters"
            return RGB(
                int(v[1:3], 16) / 255,
                int(v[3:5], 16) / 255,
                int(v[5:7], 16) / 255,
            )
        else:
            assert isinstance(v, RGB), "invalid RGB value"
            return v

    def astuple(self) -> tuple[float, float, float]:
        return (self.red, self.green, self.blue)

    def __repr__(self) -> str:
        return f"RGB({self.red:.3f}, {self.green:.3f}, {self.blue:.3f})"

    # This method cannot be defined in the class body, as it would cause a
    # circular import. The implementation is patched into the class
    # in the `style` module.
    if TYPE_CHECKING:  # pragma: no cover
        from .style import Style, StyleLike

        def __or__(self, _: StyleLike, /) -> Style:
            ...

        def __ror__(self, _: HexColor, /) -> Style:
            ...

    def __iter__(self) -> Iterator[float]:
        yield self.red
        yield self.green
        yield self.blue


red = RGB(1, 0, 0)
lime = RGB(0, 1, 0)
blue = RGB(0, 0, 1)
black = RGB(0, 0, 0)
white = RGB(1, 1, 1)
yellow = RGB(1, 1, 0)
magenta = RGB(1, 0, 1)
cyan = RGB(0, 1, 1)


@add_slots
@dataclass(frozen=True)
class dictget(Generic[T, U]):
    _map: Mapping[T, U]
    default: U

    def __call__(self, k: T) -> U:
        try:
            return self._map[k]
        except KeyError:
            return self.default


# The copious overloads are to enable mypy to
# deduce the proper callable types -- up to a limit.


T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")
T9 = TypeVar("T9")


@overload
def pipe() -> Callable[[T1], T1]:
    ...


@overload  # noqa: F811
def pipe(__f1: Callable[[T1], T2]) -> Callable[[T1], T2]:
    ...


@overload  # noqa: F811
def pipe(
    __f1: Callable[[T1], T2], __f2: Callable[[T2], T3]
) -> Callable[[T1], T3]:
    ...


@overload  # noqa: F811
def pipe(
    __f1: Callable[[T1], T2],
    __f2: Callable[[T2], T3],
    __f3: Callable[[T3], T4],
) -> Callable[[T1], T4]:
    ...


@overload  # noqa: F811
def pipe(
    __f1: Callable[[T1], T2],
    __f2: Callable[[T2], T3],
    __f3: Callable[[T3], T4],
    __f4: Callable[[T4], T5],
) -> Callable[[T1], T5]:
    ...


@overload  # noqa: F811
def pipe(
    __f1: Callable[[T1], T2],
    __f2: Callable[[T2], T3],
    __f3: Callable[[T3], T4],
    __f4: Callable[[T4], T5],
    __f5: Callable[[T5], T6],
) -> Callable[[T1], T6]:
    ...


@overload  # noqa: F811
def pipe(
    __f1: Callable[[T1], T2],
    __f2: Callable[[T2], T3],
    __f3: Callable[[T3], T4],
    __f4: Callable[[T4], T5],
    __f5: Callable[[T5], T6],
    __f6: Callable[[T6], T7],
) -> Callable[[T1], T7]:
    ...


@overload  # noqa: F811
def pipe(
    __f1: Callable[[T1], T2],
    __f2: Callable[[T2], T3],
    __f3: Callable[[T3], T4],
    __f4: Callable[[T4], T5],
    __f5: Callable[[T5], T6],
    __f6: Callable[[T6], T7],
    __f7: Callable[[T7], T8],
) -> Callable[[T1], T8]:
    ...


@overload  # noqa: F811
def pipe(
    __f1: Callable[[Any], Any],
    __f2: Callable[[Any], Any],
    __f3: Callable[[Any], Any],
    __f4: Callable[[Any], Any],
    __f5: Callable[[Any], Any],
    __f6: Callable[[Any], Any],
    __f7: Callable[[Any], Any],
    *__fn: Callable[[Any], Any],
) -> Callable[[Any], Any]:
    ...


def pipe(*__fs: Any) -> Any:  # noqa: F811
    """Create a new callable by piping several in succession
    Example
    -------
    >>> fn = pipe(float, lambda x: x / 4, int)
    >>> fn('9.3')
    2
    Note
    ----
    * Type checking is supported up to 7 functions,
      due to limitations of the Python type system.
    """
    return __pipe(__fs)


@dataclass(frozen=True, repr=False)
class __pipe:
    __slots__ = ("_functions",)
    _functions: tuple[Callable[[Any], Any], ...]

    def __call__(self, value: Any) -> Any:
        for f in self._functions:
            value = f(value)
        return value


T_contra = TypeVar("T_contra", contravariant=True)
T_co = TypeVar("T_co", covariant=True)


# shortcut for Callable[[T_contra], T_co]. Necessary for typing
# dataclass fields, as Callable is interpreted incorrectly.
class Func(Protocol[T_contra, T_co]):
    def __call__(self, __value: T_contra) -> T_co:
        ...



from __future__ import annotations

import enum
from collections import defaultdict
from dataclasses import dataclass, field
from io import BytesIO
from itertools import count
from operator import methodcaller
from pathlib import Path
from typing import TYPE_CHECKING, Collection, Iterable

from .. import atoms
from ..atoms import sanitize_name
from ..common import (
    Char,
    Func,
    Ordinal,
    Pt,
    add_slots,
    dictget,
    first,
    pipe,
    setattr_frozen,
)
from .common import (
    TEXTSPACE_TO_GLYPHSPACE,
    Font,
    FontID,
    GlyphPt,
    KerningTable,
    kern,
)

try:
    import fontTools.subset
    from fontTools.ttLib import TTFont
except ModuleNotFoundError:  # pragma: no cover
    HAS_FONTTOOLS = False
else:
    HAS_FONTTOOLS = True
    from ..vendor import get_kerning_pairs


OBJS_PER_EMBEDDED_FONT = 7
_GlyphName = str  # name uniquely identifying a glyph in a TTF font
_CID = int  # CharacterID. 16-bit ID uniquely identifying a character in a PDF
_CID_MAX = 0xFFFF  # character IDs are stored as 16-bit values
_REPLACEMENT_GLYPH: _GlyphName = ".notdef"
_CIDSYS_INFO = atoms.Dictionary(
    (b"Registry", atoms.LiteralStr(b"Adobe")),
    (b"Ordering", atoms.LiteralStr(b"UCS")),
    (b"Supplement", atoms.Int(0)),
)


def _utf16be_hex(c: Ordinal) -> bytes:
    if c <= 0xFFFF:
        return b"%04X" % c
    else:
        return b"%04X%04X" % (
            (c - 0x10000) >> 10 | 0xD800,
            (c & 0x3FF) | 0xDC00,
        )


if TYPE_CHECKING or HAS_FONTTOOLS:
    _SUBSET_OPTIONS = fontTools.subset.Options(
        drop_tables=[
            "BASE",
            "JSTF",
            "DSIG",
            "EBDT",
            "EBLC",
            "EBSC",
            "PCLT",
            "LTSH",
            "Feat",
            "Glat",
            "Gloc",
            "Silf",
            "Sill",
            "GDEF",
            "GSUB",
            "GPOS",
            "MATH",
            "hdmx",
        ],
        notdef_outline=True,
        recommended_glyphs=True,
    )

    @add_slots
    @dataclass(frozen=True, eq=False, repr=False)
    class Subset(Font):
        id: FontID
        ttf: TTFont
        charwidth: Func[Char, GlyphPt]
        cids: defaultdict[Ordinal, _CID]
        scale: float  # from glyph units to GlyphPt
        kerning: KerningTable | None
        spacewidth: GlyphPt = field(init=False)

        encoding_width = 2

        def __repr__(self) -> str:
            return (
                f"{type(self).__name__}({self.ttf['name'].getBestFullName()})"
            )

        def __post_init__(self) -> None:
            setattr_frozen(self, "spacewidth", self.charwidth(" "))

        @staticmethod
        def new(i: FontID, font: Path) -> Subset:
            ttf = TTFont(font)
            scale = TEXTSPACE_TO_GLYPHSPACE / ttf["head"].unitsPerEm
            return Subset(
                id=i,
                ttf=ttf,
                # FUTURE: cache calls to this function?
                charwidth=pipe(
                    ord,
                    dictget(ttf.getBestCmap(), _REPLACEMENT_GLYPH),
                    ttf["hmtx"].metrics.__getitem__,
                    first,
                    scale.__mul__,
                ),
                cids=defaultdict(count().__next__),
                scale=scale,
                kerning=(
                    pipe(dictget(kernpairs, 0), scale.__mul__)
                    if (kernpairs := get_kerning_pairs.for_font(ttf))
                    else None
                ),
            )

        def width(self, s: str) -> Pt:
            return sum(map(self.charwidth, s)) / TEXTSPACE_TO_GLYPHSPACE

        def encode(self, s: str) -> bytes:
            return (
                s.translate(self.cids)
                # Encoding by UTF-16BE is a fast way to convert from CIDs
                # to bytes. Because not all CIDs are valid codepoints,
                # we use surrogatepass.
                .encode("utf-16be", errors="surrogatepass")
            )

        def kern(
            self, s: str, /, prev: Char | None
        ) -> Iterable[tuple[int, GlyphPt]]:
            return kern(self.kerning, s, prev) if self.kerning else ()

        def charkern(self, a: Char, b: Char, /) -> GlyphPt:
            return self.kerning((a, b)) if self.kerning else 0

        def to_objects(self, obj_id: atoms.ObjectID) -> Iterable[atoms.Object]:
            # PDF only supports 16-bit character/glyph entries,
            # thus there is a theoretical limit to the number of unique
            # characters in a document per font. There are workarounds,
            # but most fonts don't even have this many glyphs and this many
            # different characters seem unlikely to occur in common text.
            # Thus, we just assert.
            assert len(self.cids) < _CID_MAX
            # According to PDF32000-1:2008 (9.6.4) font subsets
            # need to begin with an arbitrary six-letter tag.
            # The tag should be the same per font subset added by a PDF writer.
            # We choose `PDFJE` padded with an `A` at the front.
            tagged_name = atoms.Name(
                b"APDFJE+"
                # FUTURE: handle empty/duplicate name?
                + sanitize_name(self.ttf["name"].getBestFullName().encode())
            )
            sub = fontTools.subset.Subsetter(_SUBSET_OPTIONS)
            sub.populate(unicodes=self.cids)
            sub.subset(self.ttf)
            (
                descendant_id,
                unicodemap_id,
                cidsysinfo_id,
                descriptor_id,
                cid_gid_map_id,
                file_id,
            ) = range(obj_id + 1, obj_id + OBJS_PER_EMBEDDED_FONT)
            # These objects are based on PDF32000-1:2008, page 293
            yield (
                obj_id,
                atoms.Dictionary(
                    (b"Type", atoms.Name(b"Font")),
                    (b"Subtype", atoms.Name(b"Type0")),
                    (b"BaseFont", tagged_name),
                    (b"Encoding", atoms.Name(b"Identity-H")),
                    (
                        b"DescendantFonts",
                        atoms.Array((atoms.Ref(descendant_id),)),
                    ),
                    (b"ToUnicode", atoms.Ref(unicodemap_id)),
                ),
            )
            yield (
                descendant_id,
                atoms.Dictionary(
                    (b"Type", atoms.Name(b"Font")),
                    (b"Subtype", atoms.Name(b"CIDFontType2")),
                    (b"BaseFont", tagged_name),
                    (b"CIDSystemInfo", atoms.Ref(cidsysinfo_id)),
                    (b"FontDescriptor", atoms.Ref(descriptor_id)),
                    (
                        b"DW",
                        atoms.Real(
                            self.ttf["hmtx"][_REPLACEMENT_GLYPH][0]
                            * self.scale
                        ),
                    ),
                    (
                        b"W",
                        _encode_widths(
                            map(pipe(chr, self.charwidth), self.cids)
                        ),
                    ),
                    (b"CIDToGIDMap", atoms.Ref(cid_gid_map_id)),
                ),
            )
            yield (
                unicodemap_id,
                atoms.Stream(_unicode_map(self.cids.items())),
            )
            yield (cidsysinfo_id, _CIDSYS_INFO)
            yield (
                descriptor_id,
                atoms.Dictionary(
                    (b"Type", atoms.Name(b"FontDescriptor")),
                    (b"FontName", tagged_name),
                    (b"Flags", _encode_flags(self.ttf)),
                    (b"FontBBox", _encode_bbox(self.ttf, self.scale)),
                    (
                        b"ItalicAngle",
                        atoms.Int(round(self.ttf["post"].italicAngle)),
                    ),
                    (
                        b"Ascent",
                        atoms.Int(round(self.ttf["hhea"].ascent * self.scale)),
                    ),
                    (
                        b"Descent",
                        atoms.Int(
                            round(self.ttf["hhea"].descent * self.scale)
                        ),
                    ),
                    (b"CapHeight", _encode_cap_height(self.ttf, self.scale)),
                    (b"StemV", _encode_stem_v(self.ttf)),
                    (b"FontFile2", atoms.Ref(file_id)),
                ),
            )
            yield (cid_gid_map_id, _encode_cid_gid_map(self.ttf, self.cids))
            yield (file_id, _encode_ttf_with_side_effects(self.ttf))

    def _encode_bbox(f: TTFont, scale: float) -> atoms.Array:
        head = f["head"]
        return atoms.Array(
            atoms.Int(round(v * scale))
            for v in (head.xMin, head.yMin, head.xMax, head.yMax)
        )

    # See PDF 32000-1:2008, section 9.8.2
    class _FontDescriptorFlag(enum.IntFlag):
        FIXED_PITCH = 1 << 0
        SERIF = 1 << 1
        SYMBOLIC = 1 << 2
        SCRIPT = 1 << 3
        NONSYMBOLIC = 1 << 5
        ITALIC = 1 << 6
        ALL_CAP = 1 << 16
        SMALL_CAP = 1 << 17
        FORCE_BOLD = 1 << 18

    def _encode_flags(f: TTFont) -> atoms.Int:
        return atoms.Int(
            _FontDescriptorFlag.SYMBOLIC
            | (f["post"].isFixedPitch and _FontDescriptorFlag.FIXED_PITCH)
            | (bool(f["post"].italicAngle) and _FontDescriptorFlag.ITALIC)
            | (
                f["OS/2"].usWeightClass >= 600
                and _FontDescriptorFlag.FORCE_BOLD
            )
        )

    # WARNING: the act of saving fontTools' TTFont leads to changed properties
    # such as ttfont['head'].xMin. Beware of when you call this function!
    def _encode_ttf_with_side_effects(f: TTFont) -> atoms.Stream:
        tmp = BytesIO()
        f.save(tmp)
        content = tmp.getvalue()
        return atoms.Stream([content], [(b"Length1", atoms.Int(len(content)))])

    def _encode_cid_gid_map(ttf: TTFont, cids: Iterable[_CID]) -> atoms.Stream:
        # See PDF32000-1:2008 (table 117) for instructions on how this works.
        # Note that the Glyph IDs are different after subsetting!
        return atoms.Stream(
            map(
                pipe(
                    # CID -> Glyph name
                    dictget(ttf.getBestCmap() or {}, _REPLACEMENT_GLYPH),
                    # Glyph name -> Glyph ID
                    ttf.getReverseGlyphMap().__getitem__,
                    # Glyph ID -> bytes
                    methodcaller("to_bytes", 2, "big"),
                ),
                cids,
            )
        )

    # NOTE: the given widths must be ordered by continuously ascending
    # CID value from 0 onwards
    def _encode_widths(widths: Iterable[GlyphPt]) -> atoms.Array:
        # See PDF32000-1:2008 (9.7.4.3) for how character widths are encoded.
        # For now we choose to simply encode all widths sequentially.
        # FUTURE: It should be made more efficient for fixed width fonts.
        return atoms.Array(
            (atoms.Int(0), atoms.Array(map(pipe(int, atoms.Int), widths)))
        )

    def _encode_cap_height(f: TTFont, scale: float) -> atoms.Int:
        # Apparently not all TTF fonts properly have the capheight accessible.
        # We can fall back to the ascender height, which is usually slightly
        # lower. See https://fonts.google.com/knowledge/glossary/cap_height
        try:
            v = f["OS/2"].sCapHeight
        except AttributeError:
            v = f["hhea"].ascent

        return atoms.Int(round(v * scale))

    _FONTWEIGHT_TO_STEM_V_RATIO = (241 - 50) / (900 - 100)

    def _encode_stem_v(f: TTFont) -> atoms.Int:
        # See stackoverflow.com/questions/35485179.
        # The StemV value cannot be directly retrieved from TTF type fonts.
        # Common wisdom seems to be to map the font weight (100-900) to
        # reasonable StemV values (50-241).
        return atoms.Int(
            round(50 + _FONTWEIGHT_TO_STEM_V_RATIO * f["OS/2"].usWeightClass)
        )

    def _unicode_map(m: Collection[tuple[Ordinal, _CID]]) -> Iterable[bytes]:
        # The 'To Unicode' map ensures that the correct characters are copied
        # from a text selection.
        yield _TO_UNICODE_CMAP_PRE
        yield b"%i beginbfchar\n" % len(m)
        for code, cid in m:
            yield b"<%04X> <%b>\n" % (cid, _utf16be_hex(code))
        yield _TO_UNICODE_CMAP_POST

    # based on PDF32000-1:2008, page 294
    _TO_UNICODE_CMAP_PRE = b"""\
    /CIDInit /ProcSet findresource begin
    12 dict begin
    begincmap
    /CIDSystemInfo
    %b def
    /CMapName /Adobe-Identity-UCS def
    /CMapType 2 def
    1 begincodespacerange
    <0000> <FFFF>
    endcodespacerange
    """ % b"".join(
        _CIDSYS_INFO.write()
    )
    _TO_UNICODE_CMAP_POST = b"""\
    endbfchar
    endcmap
    CMapName currentdict /CMap defineresource pop
    end
    end"""
else:  # pragma: no cover
    FONTTOOLS_MISSING_EXCEPTION = NotImplementedError(
        "Embedded fonts require `fontTools` dependency. "
        "Install with pdfje[fonts]"
    )

    @add_slots
    @dataclass(frozen=True)
    class Subset(Font):
        @staticmethod
        def new(i: FontID, font: Path) -> Subset:
            raise FONTTOOLS_MISSING_EXCEPTION

        def width(self, s: str) -> Pt:
            raise FONTTOOLS_MISSING_EXCEPTION

        def encode(self, s: str) -> bytes:
            raise FONTTOOLS_MISSING_EXCEPTION

        def kern(self, s: str) -> Iterable[tuple[int, GlyphPt]]:
            raise FONTTOOLS_MISSING_EXCEPTION



from __future__ import annotations

from .common import RGB, XY, black, blue, cyan, lime, magenta, red, yellow
from .document import Document
from .layout.pages import AutoPage
from .page import Column, Page

__version__ = __import__("importlib.metadata").metadata.version(__name__)

__all__ = [
    # document & pages
    "Document",
    "Page",
    "Column",
    "AutoPage",
    # helpers
    "red",
    "lime",
    "blue",
    "black",
    "yellow",
    "magenta",
    "cyan",
    # common
    "RGB",
    "XY",
]



"Logic for breaking text into breakable units (words)."
# Note that a lot of the complexity here is due to the fact that we need to
# support changes to the text state (i.e. style), which can occur at any point
# in the text.
# Additionally, we try to keep performance reasonable by avoiding unnecessary
# copying, focussing on the most common cases, and using iterators.
# FUTURE: we could probablify this data model.
from __future__ import annotations

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from itertools import chain
from typing import ClassVar, Generator, Iterable, Iterator, Sequence, TypeVar

from pdfje.typeset.hyphens import Hyphenator

from ..atoms import Array, LiteralStr, Real
from ..common import (
    Char,
    NonEmptySequence,
    Pt,
    Streamable,
    add_slots,
    fix_abstract_properties,
    second,
)
from ..compat import pairwise
from ..fonts.common import TEXTSPACE_TO_GLYPHSPACE, Font, GlyphPt, Kern
from .state import NO_OP, Command, State

_T = TypeVar("_T")
_find_wordchars = re.compile(r"\w+").finditer


@fix_abstract_properties
class WordLike(ABC):
    __slots__ = ()

    # FUTURE: rename to syllables?
    @property
    @abstractmethod
    def boxes(self) -> Sequence[Slug | MixedSlug]:
        ...

    # FUTURE: rename to end state?
    @property
    @abstractmethod
    def state(self) -> State:
        ...

    @property
    @abstractmethod
    def width(self) -> Pt:
        ...

    @property
    @abstractmethod
    def tail(self) -> TrailingSpace | None:
        ...

    @abstractmethod
    def pre_state(self) -> State:
        ...

    @abstractmethod
    def last(self) -> Char:
        ...

    @abstractmethod
    def has_init_kern(self) -> bool:
        ...

    @abstractmethod
    def without_init_kern(self: _T) -> _T:
        ...

    @abstractmethod
    def indent(self: _T, amount: Pt) -> _T:
        ...

    @abstractmethod
    def pruned_width(self) -> Pt:
        ...

    @abstractmethod
    def prunable_space(self) -> Pt:
        ...

    @abstractmethod
    def pruned(self) -> WordLike:
        ...

    @abstractmethod
    def extend_tail(self: _T, amount: Pt) -> _T:
        ...

    @abstractmethod
    def stretch_tail(self: _T, ratio: float) -> _T:
        ...

    @abstractmethod
    def hyphenate(self, w: Pt) -> tuple[WordLike | None, WordLike]:
        ...

    @abstractmethod
    def minimal_box(self) -> tuple[WordLike, WordLike | None]:
        ...

    @abstractmethod
    def encode_into_line(
        self, line: Iterable[LiteralStr | Real]
    ) -> Generator[bytes, None, Iterable[LiteralStr | Real]]:
        ...

    def with_cmd(self, cmd: Command) -> WordLike:
        return self if cmd is NO_OP else WithCmd(self, cmd)


@add_slots
@dataclass(frozen=True)
class Slug(WordLike):
    "A fragment of text with its measured width and kerning information"
    txt: str  # non-empty
    kern: Sequence[Kern]
    width: Pt  # including kerning
    state: State = field(repr=False)

    tail: ClassVar[None] = None

    def pre_state(self) -> State:
        return self.state

    def last(self) -> Char:
        return self.txt[-1]

    def with_hyphen(self) -> Slug:
        kern = self.state.font.charkern(self.last(), "-")
        return Slug(
            self.txt + "-",
            [*self.kern, (len(self.txt), kern)] if kern else self.kern,
            self.width
            + (
                (self.state.font.charwidth("-") + kern)
                / TEXTSPACE_TO_GLYPHSPACE
            )
            * self.state.size,
            self.state,
        )

    def has_init_kern(self) -> bool:
        return bool(self.kern) and self.kern[0][0] == 0

    # FUTURE: rename?
    @staticmethod
    def new(s: str, state: State, prev: Char | None) -> Slug:
        font = state.font
        kern = list(font.kern(s, prev))
        return Slug(
            s,
            kern,
            (font.width(s) + sum(map(second, kern)) / TEXTSPACE_TO_GLYPHSPACE)
            * state.size,
            state,
        )

    def without_init_kern(self) -> Slug:
        kern = self.kern
        if kern:
            firstkern_pos, firstkern = kern[0]
            if firstkern_pos == 0:
                delta = (firstkern * self.state.size) / TEXTSPACE_TO_GLYPHSPACE
                return Slug(self.txt, kern[1:], self.width - delta, self.state)
        return self

    def indent(self, amount: Pt) -> Slug:
        # Shortcut for common case
        if not amount:
            return self

        # We only indent the first word of a line -- which never has Kerning
        # on the first letter.
        try:
            assert self.kern[0][0] != 0
        except IndexError:  # pragma: no cover
            pass
        return Slug(
            self.txt,
            [
                (0, amount / self.state.size * TEXTSPACE_TO_GLYPHSPACE),
                *self.kern,
            ],
            self.width + amount,
            self.state,
        )

    def pruned_width(self) -> Pt:
        return self.width

    def pruned(self) -> Slug:
        return self

    @property
    def boxes(self) -> Sequence[Slug | MixedSlug]:
        return (self,)

    def extend_tail(self, amount: Pt) -> Slug:
        return self

    def stretch_tail(self, ratio: float) -> Slug:
        return self

    def hyphenate(self, w: Pt, /) -> tuple[None, Slug]:
        return None, self

    # FUTURE: does this need to be in the interface?
    def minimal_box(self) -> tuple[Slug, None]:
        return self, None

    def prunable_space(self) -> Pt:
        return 0

    def to_atoms(self) -> Iterable[LiteralStr | Real]:
        return _encode_kerning(self.txt, self.kern, self.state.font)

    # FUTURE: we can be more efficient in writing multi-word strings.
    #         In the current situation they are always separated.
    def encode_into_line(
        self, line: Iterable[LiteralStr | Real]
    ) -> Generator[bytes, None, Iterable[LiteralStr | Real]]:
        return chain(line, self.to_atoms())
        # We need have one `yield` statement to turn this into a generator.
        # It doesn't matter that it will never be reached.
        yield  # type: ignore[unreachable]


@add_slots
@dataclass(frozen=True)
class MixedSlug(WordLike):
    """A fragment of text containing commands within it. For example a word
    which is partially italic."""

    segments: NonEmptySequence[tuple[Slug, Command]]
    state: State = field(repr=False)

    tail: ClassVar[None] = None

    def prunable_space(self) -> Pt:
        return 0

    def pruned(self) -> MixedSlug:
        return self

    def pruned_width(self) -> Pt:
        return self.width

    def minimal_box(self) -> tuple[WordLike, WordLike | None]:
        return self, None

    @property
    def boxes(self) -> Sequence[MixedSlug]:
        return (self,)

    def pre_state(self) -> State:
        return self.segments[0][0].state

    def last(self) -> Char:
        return self.segments[-1][0].last()

    def extend_tail(self, amount: Pt) -> MixedSlug:
        return self

    def stretch_tail(self, ratio: float) -> MixedSlug:
        return self

    def hyphenate(self, w: Pt, /) -> tuple[None, MixedSlug]:
        return None, self

    def with_hyphen(self) -> MixedSlug:
        return MixedSlug(
            (
                *self.segments,
                (
                    Slug.new("-", self.state, self.segments[-1][0].last()),
                    NO_OP,
                ),
            ),
            self.state,
        )

    def without_init_kern(self) -> MixedSlug:
        (first, cmd), *rest = self.segments
        return MixedSlug(
            ((first.without_init_kern(), cmd), *rest),
            self.state,
        )

    def has_init_kern(self) -> bool:
        return self.segments[0][0].has_init_kern()

    def indent(self, amount: Pt) -> MixedSlug:
        if amount:
            (first, cmd), *rest = self.segments
            return MixedSlug(
                ((first.indent(amount), cmd), *rest),
                self.state,
            )
        else:
            return self

    @property
    def width(self) -> Pt:
        return sum(s.width for s, _ in self.segments)

    def encode_into_line(
        self, line: Iterable[LiteralStr | Real]
    ) -> Generator[bytes, None, Iterable[LiteralStr | Real]]:
        for txt, cmd in self.segments:
            yield from render_kerned(chain(line, txt.to_atoms()))
            yield from cmd
            line = ()
        return line


def into_syllables(s: str, hyphens: Hyphenator) -> Iterable[str]:
    leftover = ""
    buffer: list[str] = []
    end = 0
    for match in _find_wordchars(s):
        leftover += s[end : match.start()]  # noqa: E203
        syllables = iter(hyphens(match.group()))
        buffer.append(leftover + next(syllables))
        buffer.extend(syllables)
        leftover = buffer.pop()
        yield from buffer
        buffer.clear()
        end = match.end()

    if leftover or end < len(s):
        yield leftover + s[end:]


@add_slots
@dataclass(frozen=True)
class Word(WordLike):
    # FUTURE: defer hyphenation until it is absolutely necessary.
    #         this would improve performance
    boxes: Sequence[Slug | MixedSlug]
    tail: TrailingSpace | None
    state: State = field(repr=False)

    def pre_state(self) -> State:
        try:
            first = self.boxes[0]
        except IndexError:
            return self.state
        return first.pre_state()

    def last(self) -> Char:
        return " " if self.tail else self.boxes[-1].last()

    def has_init_kern(self) -> bool:
        return (
            self.boxes[0].has_init_kern()
            if self.boxes
            else bool(self.tail and self.tail.kern)
        )

    @staticmethod
    def new(s: str, state: State, prev: Char | None) -> Word:
        s, tail = TrailingSpace.parse(s, state, prev)
        segments = []
        for part in into_syllables(s, state.hyphens):
            segments.append(Slug.new(part, state, prev))
            prev = part[-1]
        return Word(tuple(segments), tail, state)

    def hyphenate(self, space: Pt, /) -> tuple[Word | None, Word]:
        if len(self.boxes) < 1 or self.boxes[0].with_hyphen().width > space:
            return (None, self.without_init_kern())
        fitting = []
        it = iter(self.boxes)
        for box, next_box in pairwise(it):
            space -= box.width
            if next_box.with_hyphen().width > space:
                fitting.append(box.with_hyphen())
                break
            fitting.append(box)
        else:
            # We shouldn't reach this point because hypenation is only
            # called if it's necessary.
            raise RuntimeError("Hyphenation not necessary")

        return (
            Word(tuple(fitting), None, box.state) if fitting else None,
            Word((next_box.without_init_kern(), *it), self.tail, self.state),
        )

    def minimal_box(self) -> tuple[Word, Word | None]:
        "Split off a minimal part of the word, leaving the rest."
        if len(self.boxes) < 2:
            return (self, None)
        else:
            return (
                Word(
                    (self.boxes[0].with_hyphen(),), None, self.boxes[0].state
                ),
                Word(
                    (self.boxes[1].without_init_kern(), *self.boxes[2:]),
                    self.tail,
                    self.state,
                ),
            )

    def without_init_kern(self) -> Word:
        if self.boxes:
            return (
                Word(
                    (self.boxes[0].without_init_kern(), *self.boxes[1:]),
                    self.tail,
                    self.state,
                )
                if self.boxes[0].has_init_kern()
                else self
            )
        elif self.tail and self.tail.kern:
            return Word(
                (),
                TrailingSpace(self.tail.width_excl_kern, 0, self.tail.size),
                self.state,
            )
        else:
            return self

    def indent(self, amount: Pt) -> Word:
        if amount:
            if self.boxes:
                return Word(
                    (self.boxes[0].indent(amount), *self.boxes[1:]),
                    self.tail,
                    self.state,
                )
            assert self.tail  # Words with no boxes or tail shouldn't exist
            return Word((), self.tail.stretch(amount, self.state), self.state)
        else:
            return self

    def pruned_width(self) -> Pt:
        return sum(s.width for s in self.boxes)

    @property
    def width(self) -> Pt:
        return self.pruned_width() + (self.tail.width() if self.tail else 0)

    def extend_tail(self, amount: Pt) -> Word:
        return (
            Word(self.boxes, self.tail.stretch(amount, self.state), self.state)
            if self.tail and amount
            else self
        )

    def stretch_tail(self, ratio: float) -> Word:
        return (
            Word(
                self.boxes,
                self.tail.stretch(
                    ratio * self.tail.width_excl_kern, self.state
                ),
                self.state,
            )
            if self.tail and ratio != 1
            else self
        )

    def prunable_space(self) -> Pt:
        return self.tail.width() if self.tail else 0

    def pruned(self) -> Word:
        return Word(self.boxes, None, self.state) if self.tail else self

    def encode_into_line(
        self, line: Iterable[LiteralStr | Real]
    ) -> Generator[bytes, None, Iterable[LiteralStr | Real]]:
        for b in self.boxes:
            line = yield from b.encode_into_line(line)
        return (
            chain(line, self.tail.into_atoms(self.state))
            if self.tail
            else line
        )


@add_slots
@dataclass(frozen=True)
class WithCmd(WordLike):
    word: WordLike
    cmd: Command

    def last(self) -> Char:
        return self.word.last()

    def has_init_kern(self) -> bool:
        return self.word.has_init_kern()

    @property
    def state(self) -> State:
        return self.cmd.apply(self.word.state)

    def pre_state(self) -> State:
        return self.word.pre_state()

    @property
    def tail(self) -> TrailingSpace | None:
        return self.word.tail

    @property
    def boxes(self) -> Sequence[Slug | MixedSlug]:
        return self.word.boxes

    def without_init_kern(self) -> WithCmd:
        return WithCmd(self.word.without_init_kern(), self.cmd)

    @property
    def width(self) -> Pt:
        return self.word.width

    def pruned_width(self) -> Pt:
        return self.word.pruned_width()

    def pruned(self) -> WithCmd:
        return WithCmd(self.word.pruned(), self.cmd)

    def extend_tail(self, amount: Pt) -> WithCmd:
        return (
            WithCmd(self.word.extend_tail(amount), self.cmd)
            if self.word.tail and amount
            else self
        )

    def stretch_tail(self, ratio: float) -> WithCmd:
        return (
            WithCmd(self.word.stretch_tail(ratio), self.cmd)
            if self.word.tail
            else self
        )

    def hyphenate(self, w: Pt, /) -> tuple[WordLike | None, WordLike]:
        a, b = self.word.hyphenate(w)
        return a, b.with_cmd(self.cmd)

    def minimal_box(self) -> tuple[WordLike, WordLike | None]:
        a, b = self.word.minimal_box()
        if b is None:
            return a.with_cmd(self.cmd), None
        return a, b.with_cmd(self.cmd)

    def prunable_space(self) -> Pt:
        return self.word.prunable_space()

    def indent(self, amount: Pt) -> WithCmd:
        return WithCmd(self.word.indent(amount), self.cmd) if amount else self

    def encode_into_line(
        self, line: Iterable[LiteralStr | Real]
    ) -> Generator[bytes, None, Iterable[LiteralStr | Real]]:
        line = yield from self.word.encode_into_line(line)
        yield from render_kerned(line)
        yield from self.cmd
        return ()


@add_slots
@dataclass(frozen=True)
class TrailingSpace:
    width_excl_kern: Pt  # including kerning adjustment
    kern: GlyphPt
    size: Pt

    @staticmethod
    def parse(
        s: str, state: State, prev: Char | None
    ) -> tuple[str, TrailingSpace | None]:
        tail = None
        if s.endswith(" "):
            s = s[:-1]
            prev = s[-1] if s else prev
            tail = TrailingSpace(
                state.font.spacewidth / TEXTSPACE_TO_GLYPHSPACE * state.size,
                state.font.charkern(prev, " ") if prev else 0,
                state.size,
            )
        return s, tail

    # FUTURE: cache?
    def width(self) -> Pt:
        return (
            self.width_excl_kern
            + (self.kern * self.size) / TEXTSPACE_TO_GLYPHSPACE
        )

    def stretch(self, amount: Pt, state: State) -> TrailingSpace:
        return TrailingSpace(
            self.width_excl_kern,
            self.kern + (amount / state.size * TEXTSPACE_TO_GLYPHSPACE),
            self.size,
        )

    def into_atoms(self, s: State) -> Iterable[Real | LiteralStr]:
        if self.kern:
            yield Real(-self.kern)
        yield LiteralStr(s.font.encode(" "))


def render_kerned(content: Iterable[LiteralStr | Real]) -> Streamable:
    return chain(Array(content).write(), (b" TJ\n",))


def indent_first(ws: Iterable[WordLike], amount: Pt) -> Iterator[WordLike]:
    it = iter(ws)
    try:
        first = next(it)
    except StopIteration:
        return
    yield first.indent(amount)
    yield from it


# FUTURE: handle generic iterable
def _encode_kerning(
    txt: str, kerning: Sequence[Kern], f: Font
) -> Iterable[LiteralStr | Real]:
    encoded = f.encode(txt)
    try:
        index_prev, space = kerning[0]
    except IndexError:
        yield LiteralStr(encoded)
        return

    if index_prev == 0:  # i.e. the case where we kern before any text
        yield Real(-space)
        kerning = kerning[1:]

    index_prev = index = 0
    for index, space in kerning:
        index *= f.encoding_width
        yield LiteralStr(encoded[index_prev:index])
        yield Real(-space)
        index_prev = index

    yield LiteralStr(encoded[index:])



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator, final

from ..common import (
    RGB,
    XY,
    HexColor,
    Pt,
    Sides,
    SidesLike,
    Streamable,
    add_slots,
    black,
    setattr_frozen,
)
from ..resources import Resources
from ..style import StyleFull
from .common import Block, ColumnFill, Shaped


@final
@add_slots
@dataclass(frozen=True, init=False)
class Rule(Block):
    """A :class:`Block` that draws a horizontal line"""

    color: RGB
    margin: Sides

    def __init__(
        self,
        color: RGB | HexColor = black,
        margin: SidesLike = Sides(6, 0, 6, 0),
    ) -> None:
        setattr_frozen(self, "color", RGB.parse(color))
        setattr_frozen(self, "margin", Sides.parse(margin))

    def into_columns(
        self, _: Resources, __: StyleFull, cs: Iterator[ColumnFill]
    ) -> Iterator[ColumnFill]:
        col = next(cs)
        top, right, bottom, left = self.margin
        if (height := top + bottom) > col.height_free:
            # There is not enough room for the rule in the current column.
            # Yield the column and start a new one.
            yield col
        y = col.box.origin.y + col.height_free - top
        x = col.box.origin.x + left
        yield col.add(
            ShapedRule(
                XY(x, y),
                XY(col.box.origin.x + col.box.width - right, y),
                self.color,
                height,
            ),
        )


@add_slots
@dataclass(frozen=True)
class ShapedRule(Shaped):
    start: XY
    end: XY
    color: RGB
    height: Pt

    def render(self, _: XY, __: Pt) -> Streamable:
        yield b"%g %g m %g %g l %g %g %g RG S\n" % (
            *self.start,
            *self.end,
            *self.color,
        )



from __future__ import annotations

import abc
from dataclasses import dataclass
from itertools import chain
from operator import methodcaller
from typing import Iterable, Iterator, Literal, Sequence, final

from . import atoms
from .atoms import OBJ_ID_PAGETREE, OBJ_ID_RESOURCES
from .common import (
    XY,
    Sides,
    SidesLike,
    Streamable,
    add_slots,
    flatten,
    setattr_frozen,
)
from .resources import Resources
from .style import StyleFull
from .units import A4, Pt, inch

Rotation = Literal[0, 90, 180, 270]


class Drawing(abc.ABC):
    """Base class for all drawing operations wich can be put on
    a :class:`~pdfje.Page`."""

    __slots__ = ()

    @abc.abstractmethod
    def render(self, r: Resources, s: StyleFull, /) -> Streamable:
        ...


@final
@add_slots
@dataclass(frozen=True, init=False)
class Column:
    """A column to lay out block elements in.

    Parameters
    ----------
    origin
        The bottom left corner of the column. Can be parsed from a 2-tuple.
    width
        The width of the column. Must be larger than 0.
    height
        The height of the column.

    """

    origin: XY
    width: Pt
    height: Pt

    def __init__(
        self, origin: XY | tuple[Pt, Pt], width: Pt, height: Pt
    ) -> None:
        setattr_frozen(self, "origin", XY.parse(origin))
        setattr_frozen(self, "width", width)
        setattr_frozen(self, "height", height)
        assert self.width > 0


@add_slots
@dataclass(frozen=True)
class RenderedPage:
    rotate: Rotation
    size: XY
    stream: Streamable

    def to_atoms(self, i: atoms.ObjectID) -> Iterable[atoms.Object]:
        yield i, atoms.Dictionary(
            (b"Type", atoms.Name(b"Page")),
            (b"Parent", atoms.Ref(OBJ_ID_PAGETREE)),
            (b"MediaBox", atoms.Array(map(atoms.Real, [0, 0, *self.size]))),
            (b"Contents", atoms.Ref(i + 1)),
            (b"Resources", atoms.Ref(OBJ_ID_RESOURCES)),
            (b"Rotate", atoms.Int(self.rotate)),
        )
        yield i + 1, atoms.Stream(self.stream)


@final
@add_slots
@dataclass(frozen=True, init=False)
class Page:
    """A single page within a document. Contains drawings at given positions.

    Example
    -------

    .. code-block:: python

       from pdfje import Page, Line, Rect, Text, A5
       title_page = Page([
           Text((100, 200), "My awesome story"),
           Line((100, 100), (200, 100)),
           Rect((50, 50), width=200, height=300),
       ], size=A5)

    Parameters
    ----------
    content
        The drawings to render on the page.
    size
        The size of the page in points. Common page sizes are available
        as constants:

        .. code-block:: python

            from pdfje.units import Page, A4, A5, A6, letter, legal, tabloid

    rotate
        The rotation of the page in degrees.
    margin
        The margin around the page in points, used for layout.
        Can be a single value, or a 2, 3 or 4-tuple following the CSS
        shorthand convention. see https://www.w3schools.com/css/css_margin.asp
    columns
        The columns to use for laying out the content.
        If not given, the content is laid out in a single column
        based on the page size and margin.

    """

    content: Iterable[Drawing]
    size: XY
    rotate: Rotation
    columns: Sequence[Column]

    def __init__(
        self,
        content: Iterable[Drawing] = (),
        size: XY | tuple[Pt, Pt] = A4,
        rotate: Rotation = 0,
        margin: SidesLike = Sides.parse(inch(1)),
        columns: Sequence[Column] = (),
    ) -> None:
        size = XY.parse(size)
        setattr_frozen(self, "content", content)
        setattr_frozen(self, "rotate", rotate)
        setattr_frozen(self, "columns", columns or [_column(size, margin)])
        setattr_frozen(self, "size", size)

    def add(self, d: Drawing, /) -> Page:
        """Create a new page with the given drawing added

        Parameters
        ----------
        d
            The drawing to add to the page
        """
        return Page(
            (*self.content, d), self.size, self.rotate, columns=self.columns
        )

    def render(
        self, r: Resources, s: StyleFull, pnum: int, /
    ) -> Iterator[RenderedPage]:
        yield RenderedPage(
            self.rotate,
            self.size,
            flatten(map(methodcaller("render", r, s), self.content)),
        )

    def fill(
        self, r: Resources, s: StyleFull, extra: Iterable[bytes]
    ) -> RenderedPage:
        return RenderedPage(
            self.rotate,
            self.size,
            chain(
                flatten(map(methodcaller("render", r, s), self.content)),
                extra,
            ),
        )


def _column(page: XY, margin: SidesLike) -> Column:
    top, right, bottom, left = Sides.parse(margin)
    return Column(
        XY(left, bottom), page.x - left - right, page.y - top - bottom
    )



from __future__ import annotations

import abc
from dataclasses import dataclass, field
from itertools import chain, count
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Tuple, Union, final

from .. import atoms
from ..atoms import ASCII
from ..common import (
    Char,
    Func,
    Pos,
    Pt,
    add_slots,
    fix_abstract_properties,
    setattr_frozen,
)
from ..compat import pairwise

FontID = bytes  # unique, internal identifier assigned to a font within a PDF
GlyphPt = float  # length unit in glyph space
TEXTSPACE_TO_GLYPHSPACE = 1000  # See PDF32000-1:2008 (9.7.3)


@fix_abstract_properties
class Font(abc.ABC):
    """A specific font within a typeface"""

    __slots__ = ()

    @property
    @abc.abstractmethod
    def id(self) -> FontID:
        ...

    # It's worth caching this value, as it is used often
    @property
    @abc.abstractmethod
    def spacewidth(self) -> GlyphPt:
        ...

    @property
    @abc.abstractmethod
    def encoding_width(self) -> int:
        """The number of bytes assigned to each character when encoding"""

    @abc.abstractmethod
    def encode(self, s: str, /) -> bytes:
        ...

    @abc.abstractmethod
    def width(self, s: str, /) -> Pt:
        """The total width of the given string (excluding kerning)"""

    @staticmethod
    @abc.abstractmethod
    def charwidth(c: Char, /) -> GlyphPt:
        ...

    @abc.abstractmethod
    def kern(self, s: str, /, prev: Char | None) -> Iterable[Kern]:
        ...

    @abc.abstractmethod
    def charkern(self, a: Char, b: Char, /) -> GlyphPt:
        ...


@final
@add_slots
@dataclass(frozen=True, init=False)
class TrueType:
    """A TrueType font to be embedded in a PDF

    Parameters
    ----------
    regular
        The regular (i.e. non-bold, non-italic) .ttf file
    bold
        The bold .ttf file
    italic
        The italic .ttf file
    bold_italic
        The bold italic .ttf file

    """

    regular: Path
    bold: Path
    italic: Path
    bold_italic: Path

    def __init__(
        self,
        regular: Path | str,
        bold: Path | str,
        italic: Path | str,
        bold_italic: Path | str,
    ) -> None:
        setattr_frozen(self, "regular", Path(regular))
        setattr_frozen(self, "bold", Path(bold))
        setattr_frozen(self, "italic", Path(italic))
        setattr_frozen(self, "bold_italic", Path(bold_italic))

    # This method cannot be defined in the class body, as it would cause a
    # circular import. The implementation is patched into the class
    # in the `style` module.
    if TYPE_CHECKING:  # pragma: no cover
        from ..common import HexColor
        from ..style import Style, StyleLike

        def __or__(self, _: StyleLike, /) -> Style:
            ...

        def __ror__(self, _: HexColor, /) -> Style:
            ...

    def font(self, bold: bool, italic: bool) -> Path:
        if bold:
            return self.bold_italic if italic else self.bold
        else:
            return self.italic if italic else self.regular


@final
@add_slots
@dataclass(frozen=True, repr=False)
class BuiltinTypeface:
    """A typeface that is built into the PDF renderer."""

    regular: BuiltinFont
    bold: BuiltinFont
    italic: BuiltinFont
    bold_italic: BuiltinFont

    # This method cannot be defined in the class body, as it would cause a
    # circular import. The implementation is patched into the class
    # in the `style` module.
    if TYPE_CHECKING:  # pragma: no cover
        from ..common import HexColor
        from ..style import Style, StyleLike

        def __or__(self, _: StyleLike, /) -> Style:
            ...

        def __ror__(self, _: HexColor, /) -> Style:
            ...

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.regular.name.decode()})"

    def font(self, bold: bool, italic: bool) -> BuiltinFont:
        if bold:
            return self.bold_italic if italic else self.bold
        else:
            return self.italic if italic else self.regular


Typeface = Union[BuiltinTypeface, TrueType]


@final
@add_slots
@dataclass(frozen=True, eq=False)
class BuiltinFont(Font):
    name: ASCII
    id: FontID
    charwidth: Func[Char, GlyphPt] = field(repr=False)
    kerning: KerningTable | None = field(repr=False)
    spacewidth: Pt = field(init=False, repr=False)

    encoding_width = 1

    def __post_init__(self) -> None:
        setattr_frozen(self, "spacewidth", self.charwidth(" "))

    def width(self, s: str) -> Pt:
        return sum(map(self.charwidth, s)) / TEXTSPACE_TO_GLYPHSPACE

    @staticmethod
    def encode(s: str) -> bytes:
        # FUTURE: normalize unicode to allow better unicode representation
        return s.encode("cp1252", errors="replace")

    def kern(self, s: str, /, prev: Char | None) -> Iterable[Kern]:
        return kern(self.kerning, s, prev) if self.kerning else ()

    def charkern(self, a: Char, b: Char) -> GlyphPt:
        return self.kerning((a, b)) if self.kerning else 0

    def to_resource(self) -> atoms.Dictionary:
        return atoms.Dictionary(
            (b"Type", atoms.Name(b"Font")),
            (b"Subtype", atoms.Name(b"Type1")),
            (b"BaseFont", atoms.Name(self.name)),
            (b"Encoding", atoms.Name(b"WinAnsiEncoding")),
        )


KerningTable = Func[Tuple[Char, Char], GlyphPt]
Kern = Tuple[Pos, GlyphPt]


def kern(
    table: KerningTable,
    s: str,
    prev: Char | None,
) -> Iterable[Kern]:
    for i, pair in zip(
        count(not prev),
        pairwise(chain(prev, s) if prev else s),
    ):
        if space := table(pair):
            yield (i, space)



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Callable, Iterable, Iterator, Sequence

from pdfje.typeset.words import WordLike

from ..common import (
    XY,
    Align,
    Pt,
    Streamable,
    add_slots,
    fix_abstract_properties,
)
from ..layout.common import Shaped  # FUTURE: fix this near-circular dependency
from .state import State


@add_slots
@dataclass(frozen=True)
class ShapedText(Shaped):
    lines: Sequence[Line]
    lead: Pt
    align: Align
    height: Pt

    def render(self, pos: XY, width: Pt) -> Iterator[bytes]:
        return render_text(
            pos, self.pre_state(), width, self.lines, self.lead, self.align
        )

    def end_state(self) -> State | None:
        # this slightly convoluted way takes into account that lines
        # may (in rare cases) be empty
        return next(
            (w.state for s in reversed(self.lines) for w in reversed(s.words)),
            None,
        )

    def pre_state(self) -> State | None:
        # this slightly convoluted way takes into account that lines
        # may (in rare cases) be empty
        return next(
            (wd.state for ln in self.lines for wd in ln.words),
            None,
        )


def render_text(
    pos: XY,
    state: State | None,
    prev_width: Pt,
    lines: Iterable[Line],
    lead: Pt,
    align: Align,
) -> Iterator[bytes]:
    yield b"BT\n%g %g Td\n" % pos.astuple()
    yield from state or ()
    yield from _pick_renderer(align.value)(lines, lead, prev_width)
    yield b"ET\n"


@fix_abstract_properties
class Line(Streamable):
    __slots__ = ()

    @property
    @abc.abstractmethod
    def words(self) -> Sequence[WordLike]:
        ...

    @property
    @abc.abstractmethod
    def width(self) -> Pt:
        ...


def _render_left(lines: Iterable[Line], lead: Pt, _: Pt) -> Iterator[bytes]:
    yield b"%g TL\n" % lead
    for ln in lines:
        yield b"T*\n"
        yield from ln


def _render_centered(
    lines: Iterable[Line], lead: Pt, prev_width: Pt
) -> Iterator[bytes]:
    for ln in lines:
        yield b"%g %g TD\n" % ((prev_width - ln.width) / 2, -lead)
        yield from ln
        prev_width = ln.width


def _render_right(
    lines: Iterable[Line], lead: Pt, prev_width: Pt
) -> Iterator[bytes]:
    for ln in lines:
        yield b"%g %g TD\n" % ((prev_width - ln.width), -lead)
        yield from ln
        prev_width = ln.width


_pick_renderer: Callable[
    [int], Callable[[Iterable[Line], Pt, Pt], Iterable[bytes]]
] = [
    _render_left,
    _render_centered,
    _render_right,
    _render_left,  # justified lines are already stretched, so left-align.
].__getitem__



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from itertools import chain, count
from typing import Callable, Iterable, Iterator, final

from ..common import add_slots, always, flatten, setattr_frozen
from ..page import Page, RenderedPage
from ..resources import Resources
from ..style import StyleFull
from .common import Block, PageFill, fill_pages
from .paragraph import Paragraph


@final
@add_slots
@dataclass(frozen=True, init=False)
class AutoPage:
    """Automatically lays out content on multiple pages.

    Parameters
    ----------
    content: ~typing.Iterable[~pdfje.Block | str] | ~pdfje.Block | str
        The content to lay out on the pages. Can be parsed from single string
        or block.
    template: ~pdfje.Page | ~typing.Callable[[int], ~pdfje.Page]
        A page to use as a template for the layout. If a callable is given,
        it is called with the page number as the only argument to generate
        the page. Defaults to the default :class:`Page`.

    """

    content: Iterable[str | Block]
    template: Callable[[int], Page]

    def __init__(
        self,
        content: str | Block | Iterable[Block | str],
        template: Page | Callable[[int], Page] = always(Page()),
    ) -> None:
        if isinstance(content, str):
            content = [Paragraph(content)]
        elif isinstance(content, Block):
            content = [content]
        setattr_frozen(self, "content", content)

        if isinstance(template, Page):
            template = always(template)
        setattr_frozen(self, "template", template)

    def render(
        self, r: Resources, s: StyleFull, pnum: int, /
    ) -> Iterator[RenderedPage]:
        pages: Iterator[PageFill] = map(
            PageFill.new, map(self.template, count(pnum))
        )
        for block in map(_as_block, self.content):
            pages, filled = fill_pages(
                pages, partial(block.into_columns, r, s)
            )
            for p in filled:
                yield p.base.fill(r, s, flatten(p.done))

        last = next(pages)
        yield last.base.fill(r, s, flatten(chain(last.done, last.todo)))


def _as_block(b: str | Block) -> Block:
    return Paragraph(b) if isinstance(b, str) else b



from __future__ import annotations

import os
from dataclasses import dataclass
from itertools import count, islice
from pathlib import Path
from typing import IO, Iterable, Iterator, final, overload

from . import atoms
from .atoms import OBJ_ID_PAGETREE, OBJ_ID_RESOURCES
from .common import add_slots, flatten, setattr_frozen
from .layout import Block, Paragraph
from .layout.pages import AutoPage
from .page import Page
from .resources import Resources
from .style import Style, StyleFull, StyleLike

_OBJ_ID_FIRST_PAGE: atoms.ObjectID = OBJ_ID_RESOURCES + 1
_OBJS_PER_PAGE = 2


@final
@add_slots
@dataclass(frozen=True, init=False)
class Document:
    """a PDF Document

    Parameters
    ----------

    content
        The content of the document.
    style
        Change the default style of the document.

    Examples
    --------

    Below are some examples of creating documents.

    >>> Document()  # the minimal PDF -- one empty page
    >>> Document("hello world")  # a document with pages of text
    >>> Document([  # document with explicit pages
    ...     Page(...),
    ...     AutoPage([LOREM_IPSUM, ZEN_OF_PYTHON]),
    ...     Page(),
    ... ])


    note
    ----
    A document must contain at least one page to be valid
    """

    pages: Iterable[Page | AutoPage]
    style: Style

    def __init__(
        self,
        content: Iterable[Page | AutoPage] | str | Block | None = None,
        style: StyleLike = Style.EMPTY,
    ) -> None:
        if content is None:
            content = [Page()]
        elif isinstance(content, str):
            content = [AutoPage([Paragraph(content)])]
        elif isinstance(content, Block):
            content = [AutoPage([content])]

        setattr_frozen(self, "pages", content)
        setattr_frozen(self, "style", Style.parse(style))

    @overload
    def write(self) -> Iterator[bytes]:
        ...

    @overload
    def write(self, target: os.PathLike[str] | str | IO[bytes]) -> None:
        ...

    def write(  # type: ignore[return]
        self, target: os.PathLike[str] | str | IO[bytes] | None = None
    ) -> Iterator[bytes] | None:
        """Write the document to a given target. If no target is given,
        outputs the binary PDF content iteratively. See examples below.

        Parameters
        ----------
        target: ~os.PathLike | str | ~typing.IO[bytes] | None
            The target to write to. If not given, the PDF content is returned
            as an iterator.

        Returns
        -------
        ~typing.Iterator[bytes] | None

        Examples
        --------

        String, :class:`~pathlib.Path`, or :class:`~os.PathLike` target:

        >>> doc.write("myfolder/foo.pdf")
        >>> doc.write(Path.home() / "documents/foo.pdf")

        Files and file-like objects:

        >>> with open("my/file.pdf", 'wb') as f:
        ...     doc.write(f)
        >>> doc.write(b:= BytesIO())

        Iterator output is useful for streaming PDF contents. Below is
        an example of an HTTP request using the ``httpx`` library.

        >>> httpx.post("https://mysite.foo/upload", content=doc.write(),
        ...            headers={"Content-Type": "application/pdf"})
        """
        if target is None:
            return self._write_iter()
        elif isinstance(target, (str, os.PathLike)):
            self._write_to_path(Path(os.fspath(target)))
        else:  # i.e. IO[bytes]
            target.writelines(self._write_iter())

    def _write_iter(self) -> Iterator[bytes]:
        return atoms.write(_doc_objects(self.pages, self.style.setdefault()))

    def _write_to_path(self, p: Path) -> None:
        with p.open("wb") as wfile:
            wfile.writelines(self.write())


def _doc_objects(
    items: Iterable[Page | AutoPage], style: StyleFull
) -> Iterator[atoms.Object]:
    res = Resources()
    obj_id = pagenum = 0
    # FUTURE: the scoping of `pagenum` is a bit tricky here. Find a better
    #         way to do this -- or add a specific test.
    for pagenum, obj_id, page in zip(
        count(1),
        count(_OBJ_ID_FIRST_PAGE, step=_OBJS_PER_PAGE),
        flatten(p.render(res, style, pagenum + 1) for p in items),
    ):
        yield from page.to_atoms(obj_id)

    if not pagenum:
        raise RuntimeError(
            "Cannot write PDF document without at least one page"
        )
    first_font_id = obj_id + _OBJS_PER_PAGE

    yield from res.to_objects(first_font_id)
    yield from _write_headers(
        (obj_id - _OBJ_ID_FIRST_PAGE) // _OBJS_PER_PAGE + 1,
        res.to_atoms(first_font_id),
    )


_CATALOG_OBJ = (
    atoms.OBJ_ID_CATALOG,
    atoms.Dictionary(
        (b"Type", atoms.Name(b"Catalog")),
        (b"Pages", atoms.Ref(OBJ_ID_PAGETREE)),
    ),
)


def _write_headers(
    num_pages: int, resources: atoms.Dictionary
) -> Iterable[atoms.Object]:
    yield _CATALOG_OBJ
    yield (
        OBJ_ID_PAGETREE,
        atoms.Dictionary(
            (b"Type", atoms.Name(b"Pages")),
            (
                b"Kids",
                atoms.Array(
                    map(
                        atoms.Ref,
                        islice(
                            count(_OBJ_ID_FIRST_PAGE, step=_OBJS_PER_PAGE),
                            num_pages,
                        ),
                    )
                ),
            ),
            (b"Count", atoms.Int(num_pages)),
        ),
    )
    yield OBJ_ID_RESOURCES, resources



from __future__ import annotations

from dataclasses import dataclass, fields
from itertools import chain
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generator,
    Iterator,
    Sequence,
    TypeVar,
    Union,
    final,
)

from .common import RGB, HexColor, Pt, add_slots, setattr_frozen
from .fonts.builtins import helvetica
from .fonts.common import BuiltinTypeface, TrueType, Typeface
from .resources import Resources
from .typeset.hyphens import (
    Hyphenator,
    HyphenatorLike,
    default_hyphenator,
    parse_hyphenator,
)
from .typeset.state import (
    Chain,
    Command,
    Passage,
    SetColor,
    SetFont,
    SetHyphens,
    SetLineSpacing,
    State,
)

__all__ = ["Style", "Span", "StyleLike"]


# sentinel value for unset style attributes
class _NOT_SET:
    __slots__ = ()

    def __repr__(self) -> str:
        return "NOT_SET"


_NOTSET = _NOT_SET()


@final
@add_slots
@dataclass(frozen=True, init=False)
class Style:
    """Settings for visual style of text. All parameters are optional.

    The default style is Helvetica (regular) size 12, with line spacing 1.25

    Parameters
    ----------
    font: ~pdfje.fonts.TrueType | ~pdfje.fonts.BuiltinTypeface
        Typeface to use.
    size: float
        Size of the font, in points.
    bold: bool
        Whether to use bold font.
    italic: bool
        Whether to use italic font.
    color: ~pdfje.RGB | str
        Color of the text; can be given as a hex string (e.g. ``#ff0000``)
    line_spacing: float
        Line spacing, as a multiplier of the font size.
    hyphens: ~pyphen.Pyphen | None | ~typing.Callable[[str], ~typing.Iterable[str]]
        Hyphenation algorithm to use. Passing an explicit ``None`` disables
        hyphenation.

    Example
    -------

    Below we can see how to create a style, and how to combine them.

    >>> from pdfje import Style, RGB, times_roman
    >>> body = Style(font=times_roman, size=14, italic=True))
    >>> heading = body | Style(size=24, bold=True, color=RGB(0.5, 0, 0))
    >>> # fonts and colors can be used directly in place of styles
    >>> emphasis = times_roman | "#ff0000"
    """  # noqa: E501

    font: Typeface | None = None
    size: Pt | None = None
    bold: bool | None = None
    italic: bool | None = None
    color: RGB | None = None
    line_spacing: float | None = None
    hyphens: Hyphenator | None = None

    def __init__(
        self,
        font: Typeface | None = None,
        size: Pt | None = None,
        bold: bool | None = None,
        italic: bool | None = None,
        color: RGB | tuple[float, float, float] | HexColor | None = None,
        line_spacing: float | None = None,
        hyphens: HyphenatorLike | _NOT_SET = _NOTSET,
    ) -> None:
        setattr_frozen(self, "font", font)
        setattr_frozen(self, "size", size)
        setattr_frozen(self, "bold", bold)
        setattr_frozen(self, "italic", italic)
        setattr_frozen(self, "line_spacing", line_spacing)
        setattr_frozen(self, "color", color and RGB.parse(color))
        setattr_frozen(
            self,
            "hyphens",
            None
            if isinstance(hyphens, _NOT_SET)
            else parse_hyphenator(hyphens),
        )

    # Use this instead of replace() to avoid triggering __init__.
    def _evolve(self, **kwargs: object) -> Style:
        attrs = {f.name: getattr(self, f.name) for f in fields(self)}
        attrs.update(kwargs)
        new = Style.__new__(Style)
        for k, v in attrs.items():
            setattr_frozen(new, k, v)
        return new

    def __or__(self, other: StyleLike, /) -> Style:
        if isinstance(other, Style):
            return self._evolve(
                font=other.font or self.font,
                size=_fallback(other.size, self.size),
                bold=_fallback(other.bold, self.bold),
                italic=_fallback(other.italic, self.italic),
                color=other.color or self.color,
                line_spacing=_fallback(other.line_spacing, self.line_spacing),
                hyphens=_fallback(other.hyphens, self.hyphens),
            )
        elif isinstance(other, (TrueType, BuiltinTypeface)):
            return self._evolve(font=other)
        elif isinstance(other, str):
            return self._evolve(color=RGB.parse(other))
        elif isinstance(other, RGB):
            return self._evolve(color=other)
        else:
            return NotImplemented  # type: ignore[unreachable]

    def __ror__(self, other: HexColor, /) -> Style:
        return Style(color=RGB.parse(other)) | self

    def __repr__(self) -> str:
        field_reprs = [
            (f.name, v)
            for f in fields(self)
            if (v := getattr(self, f.name)) is not None
        ]
        return (
            f"Style({', '.join(f'{k}={v!r}' for k, v in field_reprs)})"
            if field_reprs
            else "Style.EMPTY"
        )

    @staticmethod
    def parse(s: StyleLike) -> Style:
        if isinstance(s, Style):
            return s
        elif isinstance(s, RGB):
            return Style(color=s)
        elif isinstance(s, (TrueType, BuiltinTypeface)):
            return Style(font=s)
        elif isinstance(s, str) and s.startswith("#"):  # type: ignore
            return Style(color=RGB.parse(s))
        else:
            raise TypeError(f"Cannot parse style from {s!r}")

    def diff(self, r: Resources, base: StyleFull) -> Iterator[Command]:
        if (
            _differs(self.bold, base.bold)
            or _differs(self.italic, base.italic)
            or _differs(self.font, base.font)
            or _differs(self.size, base.size)
        ):
            yield SetFont(
                r.font(
                    self.font or base.font,
                    _fallback(self.bold, base.bold),
                    _fallback(self.italic, base.italic),
                ),
                _fallback(self.size, base.size),
            )
        if _differs(self.color, base.color):
            yield SetColor(self.color)
        if _differs(self.line_spacing, base.line_spacing):
            yield SetLineSpacing(self.line_spacing)
        if _differs(self.hyphens, base.hyphens):
            yield SetHyphens(self.hyphens)

    def setdefault(self) -> StyleFull:
        return StyleFull.DEFAULT | self

    EMPTY: ClassVar[Style]


StyleLike = Union[Style, RGB, Typeface, HexColor]
Style.EMPTY = Style()

bold = Style(bold=True)
"""Shortcut for bold style."""
italic = Style(italic=True)
"""Shortcut for italic style."""
regular = Style(bold=False, italic=False)
"""Shortcut for regular (non-bold or italic) style."""


@add_slots
@dataclass(frozen=True)
class StyleFull:
    font: Typeface
    size: Pt
    bold: bool
    italic: bool
    color: RGB
    line_spacing: float
    hyphens: Hyphenator

    def __or__(self, s: Style, /) -> StyleFull:
        return StyleFull(
            s.font or self.font,
            _fallback(s.size, self.size),
            _fallback(s.bold, self.bold),
            _fallback(s.italic, self.italic),
            s.color or self.color,
            _fallback(s.line_spacing, self.line_spacing),
            s.hyphens or self.hyphens,
        )

    def as_state(self, res: Resources) -> State:
        return State(
            res.font(self.font, self.bold, self.italic),
            self.size,
            self.color,
            self.line_spacing,
            self.hyphens,
        )

    def diff(self, res: Resources, base: StyleFull) -> Iterator[Command]:
        if not (
            self.bold == base.bold
            and self.italic == base.italic
            and self.font == base.font
            and self.size == base.size
        ):
            yield SetFont(
                res.font(self.font, self.bold, self.italic),
                self.size,
            )
        if self.color != base.color:
            yield SetColor(self.color)

        if self.line_spacing != base.line_spacing:
            yield SetLineSpacing(self.line_spacing)

        if self.hyphens != base.hyphens:
            yield SetHyphens(self.hyphens)

    DEFAULT: ClassVar[StyleFull]


StyleFull.DEFAULT = StyleFull(
    helvetica, 12, False, False, RGB(0, 0, 0), 1.25, default_hyphenator
)


_T = TypeVar("_T")


def _fallback(a: _T | None, b: _T) -> _T:
    return b if a is None else a


if TYPE_CHECKING:  # pragma: no cover
    from typing import TypeGuard

    def _differs(a: _T | None, b: _T) -> TypeGuard[_T]:
        return a is not None and a != b

else:

    def _differs(a: _T | None, b: _T) -> bool:
        return a is not None and a != b


class StyledMixin:
    "A mixin for shared behavior of styled text classes"
    __slots__ = ()
    content: Sequence[str | Span]
    style: Style

    def flatten(
        self,
        r: Resources,
        base: StyleFull,
        todo: Iterator[Command] = iter(()),
    ) -> Generator[Passage, None, Iterator[Command]]:
        todo = chain(todo, self.style.diff(r, base))
        newbase = base | self.style
        for item in self.content:
            if isinstance(item, str):
                yield Passage(Chain.squash(todo), item)
            else:
                todo = yield from item.flatten(r, newbase, todo)
        return chain(todo, base.diff(r, newbase))


@final
@add_slots
@dataclass(frozen=True, init=False)
class Span(StyledMixin):
    """A fragment of text with a style.

    Parameters
    ----------
    content: str | Span | ~typing.Sequence[str | Span]
        The text to render. Can be a string, or a nested :class:`~pdfje.Span`.
    style
        The style to render the text with.
        See :ref:`tutorial<style>` for more details.

    Examples
    --------

    .. code-block:: python

        from pdfje.style import Span, Style, bold
        from pdfje.fonts import times_roman

        # A simple span
        Span("Hello, world!", Style(size=24, color="#ff0000"))
        # A nested span
        Span([
            "Beautiful is ",
            Span("better", helvetica | bold),
            " than ugly.",
        ], style=times_roman)
    """

    content: Sequence[str | Span]
    style: Style

    def __init__(
        self,
        content: str | Span | Sequence[str | Span],
        style: StyleLike = Style.EMPTY,
    ):
        if isinstance(content, (str, Span)):
            content = [content]
        setattr_frozen(self, "content", content)
        setattr_frozen(self, "style", Style.parse(style))


# The implementation of these operators are patched onto existing classes here
# to avoid circular imports.
if not TYPE_CHECKING:  # pragma: no branch

    def _typeface__or__(self, other: StyleLike, /) -> Style:
        return Style(font=self) | other

    def _typeface__ror__(self, other: HexColor, /) -> Style:
        return Style(font=self) | RGB.parse(other)

    BuiltinTypeface.__or__ = _typeface__or__
    BuiltinTypeface.__ror__ = _typeface__ror__
    TrueType.__or__ = _typeface__or__
    TrueType.__ror__ = _typeface__ror__

    def _rgb__or__(self, other: StyleLike, /) -> Style:
        return Style(color=self) | other

    def _rgb__ror__(self, other: HexColor, /) -> Style:
        return Style(color=self)

    RGB.__or__ = _rgb__or__
    RGB.__ror__ = _rgb__ror__



from __future__ import annotations

from bisect import bisect
from dataclasses import dataclass
from math import inf
from typing import Callable, ClassVar, Iterable, Iterator, Protocol, Sequence

from pdfje.fonts.common import TEXTSPACE_TO_GLYPHSPACE

from ..atoms import LiteralStr, Real
from ..common import XY, Align, Pt, add_slots, peek
from ..compat import cache
from .knuth_plass import Box, Break, NoFeasibleBreaks, optimum_fit
from .layout import Line as _Line
from .layout import ShapedText
from .state import NO_OP, Command
from .words import WithCmd, Word, WordLike, render_kerned

_STRETCH_RATIO = 0.5
_SHRINK_RATIO = 1 / 3


class Parameters(Protocol):
    tolerance: float
    hyphen_penalty: float
    consecutive_hyphen_penalty: float
    fitness_diff_penalty: float


class ColumnQueue:
    lead: Pt
    _queue: Iterator[XY]  # (infinite) queue of columns
    columns: list[XY]  # columns which have already been read from the queue
    line_counts: list[int]  # number of lines in each column, cumulative.

    __slots__ = ("lead", "_queue", "columns", "line_counts")

    def __init__(
        self, queue: Iterable[XY], lead: Pt, allow_empty: bool
    ) -> None:
        self.lead = lead
        self._queue = queue = iter(queue)
        self.columns = [nextcol := next(queue)]
        self.line_counts = [int(nextcol.y // self.lead or not allow_empty)]

    def _next_column(self) -> XY:
        col = next(self._queue)
        self.columns.append(col)
        self.line_counts.append(
            self.line_counts[-1] + (int(col.y // self.lead) or 1)
        )
        return col

    def line_length(self, i: int) -> Pt:
        try:
            return self.columns[bisect(self.line_counts, i)].x
        except IndexError:
            while self.line_counts[-1] <= i:  # always true on first iteration
                col = self._next_column()
            return col.x

    def remove_first_column(self) -> None:
        self.columns.pop(0)
        firstcount = self.line_counts.pop(0)
        self.line_counts[:] = map(firstcount.__rsub__, self.line_counts)

    def shorten_column(self, i: int) -> None:
        self.line_counts[i:] = map((1).__rsub__, self.line_counts[i:])


@add_slots
@dataclass(frozen=True)
class Fragment:
    txt: WordLike
    on_break: WordLike | None
    cmd: Command


def _to_boxes(
    ws: Iterable[WordLike], justify: bool
) -> Iterator[tuple[Fragment, Box]]:
    width = 0.0
    shrink = 0.0
    if justify:
        stretch = 0.0
    else:
        try:
            word_first, ws = peek(iter(ws))
        except StopIteration:
            return
        # For ragged text, we need to specify a constant stretch.
        # good value is 3 times the width of a space (see knuth_plass.py)
        stretch = (
            word_first.state.size
            * (word_first.state.font.spacewidth / TEXTSPACE_TO_GLYPHSPACE)
            * 3
        )

    for word in ws:
        if isinstance(word, WithCmd):
            cmd = word.cmd
            word = word.word
        else:
            cmd = NO_OP
        for chunk in word.boxes[:-1]:
            with_hyphen = chunk.with_hyphen()
            width_new = width + chunk.width
            yield Fragment(
                chunk,
                with_hyphen,
                NO_OP,
            ), Box(
                width_new,
                stretch,
                shrink,
                width_new,
                width + with_hyphen.width,
                hyphenated=True,
                no_break=False,
            )
            width = width_new

        if word.tail:
            syllables = word.boxes[-1:]
            width += syllables[0].width if syllables else 0
            yield Fragment(
                Word(syllables, word.tail, word.state),
                syllables[0] if syllables else None,
                cmd,
            ), Box(
                width,
                stretch,
                shrink,
                # FUTURE: account for initial kern of the next word
                width_new := width + word.tail.width(),
                width,
                hyphenated=False,
                no_break=False,
            )
            # FUTURE: prevent running this every loop
            if justify:
                stretch += word.tail.width_excl_kern * _STRETCH_RATIO
                shrink += word.tail.width_excl_kern * _SHRINK_RATIO
            width = width_new
        else:  # a break made possible by non-space (e.g. a hyphen, or a dash)
            ending = word.boxes[-1]
            width += ending.width
            yield Fragment(ending, ending, cmd), Box(
                width,
                stretch,
                shrink,
                width,
                width,
                hyphenated=ending.last() in "-\N{EM DASH}",
                no_break=False,
            )


def into_boxes(
    ws: Iterable[WordLike], justify: bool
) -> tuple[Sequence[Fragment], Sequence[Box]]:
    [*result] = zip(*_to_boxes(ws, justify))
    if not result:  # i.e. there were no words
        return [], []
    return tuple(result)  # type: ignore[return-value]


@add_slots
@dataclass(frozen=True)
class Line(_Line):
    words: Sequence[WordLike]
    width: Pt

    def __iter__(self) -> Iterator[bytes]:
        content: Iterable[Real | LiteralStr] = ()
        for w in self.words:
            content = yield from w.encode_into_line(content)
        yield from render_kerned(content)

    def stretch(self, adjust: float) -> Line:
        return Line(
            [
                f.stretch_tail(
                    adjust * (_STRETCH_RATIO if adjust > 0 else _SHRINK_RATIO)
                )
                for f in self.words
            ],
            # FUTURE: we don't adjust the width here,
            # because we don't use it if the line is stretched.
            # Of course, this should be adressed in a nicer way.
            self.width,
        )

    EMPTY: ClassVar[Line]


Line.EMPTY = Line((), 0)


def _into_lines_justified(
    breaks: Sequence[Break], frags: Sequence[Fragment]
) -> Iterable[Line]:
    for br, line in zip(breaks, _into_lines_ragged(breaks, frags)):
        yield line.stretch(br.adjust)


def _into_lines_ragged(
    breaks: Sequence[Break], frags: Sequence[Fragment]
) -> Iterable[Line]:
    pos_prev = 0
    for br in breaks:
        *init, last = frags[pos_prev : br.pos]  # noqa: E203
        body = [f.txt.with_cmd(f.cmd) for f in init]
        if tail := last.on_break:
            body.append(tail.with_cmd(last.cmd))
        elif body:
            body[-1] = body[-1].with_cmd(last.cmd)
        else:
            # FUTURE: strictly speaking, we should do something here
            #         with last.cmd. However, in practice, it's always
            #         the last line, and it's almost always NO_OP.
            pass
        yield Line(body, br.width)
        pos_prev = br.pos


def find_breaks(
    boxes: Sequence[Box],
    line_length: Callable[[int], float],
    ragged: bool,
    params: Parameters,
) -> Sequence[Break]:
    assert boxes
    # OPTIMIZE: a smarter way of adjusting the tolerance
    for scale in 1, 2, 3, 5, 8, inf:
        try:
            return optimum_fit(
                boxes,
                line_length,
                params.tolerance * scale,
                params.hyphen_penalty,
                params.consecutive_hyphen_penalty,
                params.fitness_diff_penalty,
                ragged=ragged,
            )
        except NoFeasibleBreaks:
            pass
    else:  # pragma: no cover
        raise RuntimeError(
            "Error in optimal line breaking algorithm. "
            "Is the line length too small to fit single words? "
            "If not, please report this as a bug."
        )


def _lines_per_column(
    ls: Sequence[Line], counts: Iterable[int]
) -> Iterator[Sequence[Line]]:
    prev_index = 0
    for linecount in counts:
        lines_in_subpara = ls[prev_index:linecount]
        if not lines_in_subpara:
            break
        yield lines_in_subpara
        prev_index = linecount


def shape(
    ws: Iterator[WordLike],
    columns: Iterator[XY],
    allow_empty: bool,
    lead: Pt,
    avoid_orphans: bool,
    align: Align,
    params: Parameters,
) -> Iterator[ShapedText]:
    ragged = align is not Align.JUSTIFY
    into_lines = _into_lines_ragged if ragged else _into_lines_justified
    fragments, boxes = into_boxes(ws, justify=not ragged)

    col_queue = ColumnQueue(columns, lead, allow_empty)

    # if the first column has no lines, fast-forward to the next one
    if col_queue.line_counts[0] == 0:
        yield ShapedText((), lead, align, 0)
        col_queue = ColumnQueue(columns, lead, allow_empty=False)

    if not boxes:
        yield ShapedText((Line.EMPTY,), lead, align, lead)
        return

    breaks = find_breaks(boxes, cache(col_queue.line_length), ragged, params)
    lines = list(into_lines(breaks, fragments))

    # redo everything if there's an orphaned first line
    # OPTIMIZE: to prevent running the optimization twice, we could
    #           do an initial check if the first line will very *likely*
    #           be orphaned. This check can't be 100% accurate, but it
    #           would save time for obvious cases.
    if (
        avoid_orphans
        and allow_empty
        and col_queue.line_counts[0] == 1
        and len(lines) > 1
        and col_queue.line_counts[1] >= 2
    ):
        yield ShapedText((), lead, align, 0)
        col_queue.remove_first_column()
        breaks = find_breaks(
            boxes, cache(col_queue.line_length), ragged, params
        )
        lines = list(into_lines(breaks, fragments))

    grouped_lines = list(_lines_per_column(lines, col_queue.line_counts))

    # redo everything if there's an orphaned last line -- this may
    # need to happen multiple times if column widths are very different.
    if avoid_orphans:
        while (
            # we've found an orphaned last line
            len(grouped_lines[-1]) == 1
            # ...and there's enough room for another line
            and col_queue.line_counts[len(grouped_lines) - 1]
            - col_queue.line_counts[len(grouped_lines) - 2]
            > 1
            # ...and moving the previous line won't leave an orphan there
            and len(grouped_lines) > 1
            and len(grouped_lines[-2]) > 2
        ):
            col_queue.shorten_column(len(grouped_lines) - 2)
            breaks = find_breaks(
                boxes, cache(col_queue.line_length), ragged, params
            )
            lines = list(into_lines(breaks, fragments))
            grouped_lines_new = list(
                _lines_per_column(lines, col_queue.line_counts)
            )
            if (
                len(grouped_lines) == len(grouped_lines_new)
                and len(grouped_lines_new[-1]) == 1
            ):
                # Our attempt to fix the orphan failed. This can happen if
                # the last line is very long and the previous line is very
                # short. In this case, we'll just have to live with the
                # original orphan.
                break
            grouped_lines = grouped_lines_new  # FUTURE: test the need for this

    for ls in grouped_lines:
        yield ShapedText(ls, lead, align, len(ls) * lead)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Iterator, Literal, Sequence, final

from .atoms import LiteralStr, Real
from .common import (
    RGB,
    XY,
    Align,
    HexColor,
    Pt,
    Streamable,
    add_slots,
    pipe,
    setattr_frozen,
)
from .page import Drawing
from .resources import Resources
from .style import Span, Style, StyledMixin, StyleFull, StyleLike
from .typeset.layout import Line as _TextLineBase
from .typeset.layout import render_text
from .typeset.parse import into_words
from .typeset.state import Command, Passage, State, max_lead, splitlines
from .typeset.words import WordLike, render_kerned

__all__ = [
    "Circle",
    "Ellipse",
    "Line",
    "Polyline",
    "Rect",
    "Text",
    "Drawing",
]


@final
@add_slots
@dataclass(frozen=True, init=False)
class Line(Drawing):
    """A :class:`Drawing` of a straight line segment.

    Parameters
    ----------
    start
        The start point of the line. Can be parsed from a 2-tuple.
    end
        The end point of the line. Can be parsed from a 2-tuple.
    stroke: RGB | str | None
        The color -- can be parsed from a hex string (e.g. ``#ff0000``)

    """

    start: XY
    end: XY
    stroke: RGB | None

    def __init__(
        self,
        start: XY | tuple[float, float],
        end: XY | tuple[float, float],
        stroke: RGB | HexColor | None = RGB(0, 0, 0),
    ) -> None:
        setattr_frozen(self, "start", XY.parse(start))
        setattr_frozen(self, "end", XY.parse(end))
        setattr_frozen(self, "stroke", stroke and RGB.parse(stroke))

    def render(self, _: Resources, __: StyleFull, /) -> Streamable:
        yield b"%g %g m %g %g l " % (*self.start, *self.end)
        yield from _finish(None, self.stroke, False)


@final
@add_slots
@dataclass(frozen=True, init=False)
class Rect(Drawing):
    """A :class:`Drawing` of a rectangle.

    Parameters
    ----------
    origin
        The bottom left corner of the rectangle. Can be parsed from a 2-tuple.
    width
        The width of the rectangle.
    height
        The height of the rectangle.
    fill: RGB | str | None
        The fill color -- can be parsed from a hex string (e.g. ``#ff0000``)
    stroke: RGB | str | None
        The stroke color -- can be parsed from a hex string (e.g. ``#ff0000``)

    """

    origin: XY
    width: Pt
    height: Pt
    fill: RGB | None
    stroke: RGB | None

    def __init__(
        self,
        origin: XY | tuple[float, float],
        width: Pt,
        height: Pt,
        fill: RGB | HexColor | None = None,
        stroke: RGB | HexColor | None = RGB(0, 0, 0),
    ) -> None:
        setattr_frozen(self, "origin", XY.parse(origin))
        setattr_frozen(self, "width", width)
        setattr_frozen(self, "height", height)
        setattr_frozen(self, "fill", fill and RGB.parse(fill))
        setattr_frozen(self, "stroke", stroke and RGB.parse(stroke))

    def render(self, _: Resources, __: StyleFull, /) -> Streamable:
        yield b"%g %g %g %g re " % (*self.origin, self.width, self.height)
        yield from _finish(self.fill, self.stroke, False)


@final
@add_slots
@dataclass(frozen=True, init=False)
class Ellipse(Drawing):
    """A :class:`Drawing` of an ellipse.

    Parameters
    ----------
    center
        The center of the ellipse. Can be parsed from a 2-tuple.
    width
        The width of the ellipse.
    height
        The height of the ellipse.
    fill: RGB | str | None
        The fill color -- can be parsed from a hex string (e.g. ``#ff0000``)
    stroke: RGB | str | None
        The stroke color -- can be parsed from a hex string (e.g. ``#ff0000``)

    """

    center: XY
    width: Pt
    height: Pt
    fill: RGB | None
    stroke: RGB | None

    def __init__(
        self,
        center: XY | tuple[float, float],
        width: Pt,
        height: Pt,
        fill: RGB | HexColor | None = None,
        stroke: RGB | HexColor | None = RGB(0, 0, 0),
    ) -> None:
        setattr_frozen(self, "center", XY.parse(center))
        setattr_frozen(self, "width", width)
        setattr_frozen(self, "height", height)
        setattr_frozen(self, "fill", fill and RGB.parse(fill))
        setattr_frozen(self, "stroke", stroke and RGB.parse(stroke))

    def render(self, _: Resources, __: StyleFull, /) -> Streamable:
        return _ellipse(
            self.center, self.width, self.height, self.fill, self.stroke
        )


@final
@add_slots
@dataclass(frozen=True, init=False)
class Circle(Drawing):
    """A :class:`Drawing` of a circle.

    Parameters
    ----------
    center
        The center of the circle. Can be parsed from a 2-tuple.
    radius
        The radius of the circle.
    fill: RGB | str | None
        The fill color -- can be parsed from a hex string (e.g. ``#ff0000``)
    stroke: RGB | str | None
        The stroke color -- can be parsed from a hex string (e.g. ``#ff0000``)

    """

    center: XY
    radius: Pt
    fill: RGB | None
    stroke: RGB | None

    def __init__(
        self,
        center: XY | tuple[float, float],
        radius: Pt,
        fill: RGB | HexColor | None = None,
        stroke: RGB | HexColor | None = RGB(0, 0, 0),
    ) -> None:
        setattr_frozen(self, "center", XY.parse(center))
        setattr_frozen(self, "radius", radius)
        setattr_frozen(self, "fill", fill and RGB.parse(fill))
        setattr_frozen(self, "stroke", stroke and RGB.parse(stroke))

    def render(self, _: Resources, __: StyleFull, /) -> Streamable:
        width = self.radius * 2
        return _ellipse(self.center, width, width, self.fill, self.stroke)


def _finish(fill: RGB | None, stroke: RGB | None, close: bool) -> Streamable:
    if fill and stroke:
        yield b"%g %g %g rg %g %g %g RG " % (*fill, *stroke)
        yield b"b\n" if close else b"B\n"
    elif fill:
        yield b"%g %g %g rg f\n" % fill.astuple()
    elif stroke:
        yield b"%g %g %g RG " % stroke.astuple()
        yield b"s\n" if close else b"S\n"
    else:
        yield b"n\n"


# based on https://stackoverflow.com/questions/2172798
def _ellipse(
    center: XY, w: Pt, h: Pt, fill: RGB | None, stroke: RGB | None
) -> Streamable:
    x, y = center - (w / 2, h / 2)
    kappa = 0.5522848
    ox = (w / 2) * kappa
    oy = (h / 2) * kappa
    xe = x + w
    ye = y + h
    xm = x + w / 2
    ym = y + h / 2
    yield b"%g %g m " % (x, ym)
    yield b"%g %g %g %g %g %g c " % (
        x,
        ym - oy,
        xm - ox,
        y,
        xm,
        y,
    )
    yield b"%g %g %g %g %g %g c " % (
        xm + ox,
        y,
        xe,
        ym - oy,
        xe,
        ym,
    )
    yield b"%g %g %g %g %g %g c " % (
        xe,
        ym + oy,
        xm + ox,
        ye,
        xm,
        ye,
    )
    yield b"%g %g %g %g %g %g c " % (
        xm - ox,
        ye,
        x,
        ym + oy,
        x,
        ym,
    )
    yield from _finish(fill, stroke, False)


@final
@add_slots
@dataclass(frozen=True, init=False)
class Polyline(Drawing):
    """A :class:`Drawing` of a polyline.

    Parameters
    ----------
    points
        The points of the polyline. Can be parsed from a 2-tuple.
    close
        Whether to close the polyline.
    fill: RGB | str | None
        The fill color -- can be parsed from a hex string (e.g. ``#ff0000``)
    stroke: RGB | str | None
        The stroke color -- can be parsed from a hex string (e.g. ``#ff0000``)

    """

    points: Iterable[XY | tuple[float, float]]
    close: bool = False
    fill: RGB | None = None
    stroke: RGB | None = RGB(0, 0, 0)

    def __init__(
        self,
        points: Iterable[XY | tuple[float, float]],
        close: bool = False,
        fill: RGB | HexColor | None = None,
        stroke: RGB | HexColor | None = RGB(0, 0, 0),
    ) -> None:
        setattr_frozen(self, "points", points)
        setattr_frozen(self, "close", close)
        setattr_frozen(self, "fill", fill and RGB.parse(fill))
        setattr_frozen(self, "stroke", stroke and RGB.parse(stroke))

    def render(self, _: Resources, __: StyleFull, /) -> Streamable:
        it = iter(self.points)
        try:
            yield b"%g %g m " % next(it)
        except StopIteration:
            return
        yield from map(pipe(tuple, b"%g %g l ".__mod__), it)
        yield from _finish(self.fill, self.stroke, self.close)


@final
@add_slots
@dataclass(frozen=True, init=False)
class Text(Drawing, StyledMixin):
    """A :class:`Drawing` of text at the given location (not wrapped)

    Parameters
    ----------
    loc
        The location of the text. Can be parsed from a 2-tuple.
    content: str | Span | ~typing.Sequence[str | Span]
        The text to render. Can be a string, or a nested :class:`~pdfje.Span`.
    style
        The style to apply to the text.
    align
        The horizontal alignment of the text.
    """

    loc: XY
    content: Sequence[str | Span]
    style: Style
    align: Align

    def __init__(
        self,
        loc: XY | tuple[float, float],
        content: str | Span | Sequence[str | Span],
        style: StyleLike = Style.EMPTY,
        align: Align | Literal["left", "center", "right"] = Align.LEFT,
    ) -> None:
        if isinstance(content, (str, Span)):
            content = [content]
        setattr_frozen(self, "loc", XY.parse(loc))
        setattr_frozen(self, "content", content)
        setattr_frozen(self, "style", Style.parse(style))
        setattr_frozen(self, "align", Align.parse(align))
        # FUTURE: a more elegant way to do this
        if self.align is Align.JUSTIFY:
            raise NotImplementedError(
                "Justified alignment not implemented for explicitly "
                "positioned text."
            )

    def render(self, r: Resources, s: StyleFull, /) -> Streamable:
        state = s.as_state(r)
        passages = list(self.flatten(r, s))
        return render_text(
            self.loc,
            state,
            0,
            list(into_lines(splitlines(passages), state)),
            max_lead(passages, state),
            self.align,
        )


@add_slots
@dataclass(frozen=True)
class _TextLine(_TextLineBase):
    cmd: Command
    words: tuple[WordLike, ...]
    width: Pt

    def __iter__(self) -> Iterator[bytes]:
        yield from self.cmd
        content: Iterable[Real | LiteralStr] = ()
        for w in self.words:
            content = yield from w.encode_into_line(content)
        yield from render_kerned(content)


def into_lines(
    split: Iterable[Iterable[Passage]], state: State
) -> Iterator[_TextLine]:
    for s in split:
        cmd, [*words] = into_words(s, state)
        yield _TextLine(cmd, tuple(words), sum(w.width for w in words))
        try:
            state = words[-1].state
        except IndexError:
            pass  # empty line -- no change to state



# This file has been taken from:
# github.com/adobe-type-tools/kern-dump/blob/main/getKerningPairsFromOTF.py
# under its MIT license:
#
# Copyright (c) 2013-2016 Adobe Systems Incorporated. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# The following changes have been made to the original file:
# - Removed the main function and the command line argument parsing
# - Removed the print statements
# - OTFKernReader takes a TTFont object as input instead of a path
# - added a `for_font` helper method
# - formatted with black and isort
from __future__ import annotations

from typing import Mapping

from fontTools import ttLib

kKernFeatureTag = "kern"
kGPOStableName = "GPOS"


class myLeftClass:
    def __init__(self):
        self.glyphs = []
        self.class1Record = 0


class myRightClass:
    def __init__(self):
        self.glyphs = []
        self.class2Record = 0


def collect_unique_kern_lookup_indexes(featureRecord):
    unique_kern_lookups = []
    for featRecItem in featureRecord:
        # GPOS feature tags (e.g. kern, mark, mkmk, size) of each ScriptRecord
        if featRecItem.FeatureTag == kKernFeatureTag:
            feature = featRecItem.Feature

            for featLookupItem in feature.LookupListIndex:
                if featLookupItem not in unique_kern_lookups:
                    unique_kern_lookups.append(featLookupItem)

    return unique_kern_lookups


class OTFKernReader(object):
    def __init__(self, font):
        self.font = font
        self.kerningPairs = {}
        self.singlePairs = {}
        self.classPairs = {}
        self.pairPosList = []
        self.allLeftClasses = {}
        self.allRightClasses = {}

        if kGPOStableName not in self.font:
            self.goodbye()

        else:
            self.analyzeFont()
            self.findKerningLookups()
            self.getPairPos()
            self.getSinglePairs()
            self.getClassPairs()

    def goodbye(self):
        return

    def analyzeFont(self):
        self.gposTable = self.font[kGPOStableName].table
        self.scriptList = self.gposTable.ScriptList
        self.featureList = self.gposTable.FeatureList
        self.featureCount = self.featureList.FeatureCount
        self.featureRecord = self.featureList.FeatureRecord

        self.unique_kern_lookups = collect_unique_kern_lookup_indexes(
            self.featureRecord
        )

    def findKerningLookups(self):
        if not len(self.unique_kern_lookups):
            self.goodbye()

        self.lookup_list = self.gposTable.LookupList
        self.lookups = []
        for kern_lookup_index in sorted(self.unique_kern_lookups):
            lookup = self.lookup_list.Lookup[kern_lookup_index]

            # Confirm this is a GPOS LookupType 2; or
            # using an extension table (GPOS LookupType 9):

            """
            Lookup types:
            1   Single adjustment           Adjust position of a single glyph
            2   Pair adjustment             Adjust position of a pair of glyphs
            3   Cursive attachment          Attach cursive glyphs
            4   MarkToBase attachment       Attach a combining mark to a base glyph
            5   MarkToLigature attachment   Attach a combining mark to a ligature
            6   MarkToMark attachment       Attach a combining mark to another mark
            7   Context positioning         Position one or more glyphs in context
            8   Chained Context positioning Position one or more glyphs in chained context
            9   Extension positioning       Extension mechanism for other positionings
            10+ Reserved for future use
            """

            if lookup.LookupType not in [2, 9]:
                continue
            self.lookups.append(lookup)

    def getPairPos(self):
        for lookup in self.lookups:
            for subtableItem in lookup.SubTable:
                if subtableItem.LookupType == 9:  # extension table
                    if subtableItem.ExtensionLookupType == 8:  # contextual
                        continue
                    elif subtableItem.ExtensionLookupType == 2:
                        subtableItem = subtableItem.ExtSubTable

                self.pairPosList.append(subtableItem)

                # Each glyph in this list will have a corresponding PairSet
                # which will contain all the second glyphs and the kerning
                # value in the form of PairValueRecord(s)
                # self.firstGlyphsList.extend(subtableItem.Coverage.glyphs)

    def getSinglePairs(self):
        for pairPos in self.pairPosList:
            if pairPos.Format == 1:
                # single pair adjustment

                firstGlyphsList = pairPos.Coverage.glyphs

                # This iteration is done by index so we have a way
                # to reference the firstGlyphsList:
                for ps_index, _ in enumerate(pairPos.PairSet):
                    for pairValueRecordItem in pairPos.PairSet[
                        ps_index
                    ].PairValueRecord:
                        secondGlyph = pairValueRecordItem.SecondGlyph
                        valueFormat = pairPos.ValueFormat1

                        if valueFormat == 5:  # RTL kerning
                            kernValue = "<%d 0 %d 0>" % (
                                pairValueRecordItem.Value1.XPlacement,
                                pairValueRecordItem.Value1.XAdvance,
                            )
                        elif valueFormat == 0:  # RTL pair with value <0 0 0 0>
                            kernValue = "<0 0 0 0>"
                        elif valueFormat == 4:  # LTR kerning
                            kernValue = pairValueRecordItem.Value1.XAdvance
                        else:
                            continue  # skip the rest

                        self.kerningPairs[
                            (firstGlyphsList[ps_index], secondGlyph)
                        ] = kernValue
                        self.singlePairs[
                            (firstGlyphsList[ps_index], secondGlyph)
                        ] = kernValue

    def getClassPairs(self):
        for loop, pairPos in enumerate(self.pairPosList):
            if pairPos.Format == 2:
                leftClasses = {}
                rightClasses = {}

                # Find left class with the Class1Record index="0".
                # This first class is mixed into the "Coverage" table
                # (e.g. all left glyphs) and has no class="X" property
                # that is why we have to find the glyphs in that way.

                lg0 = myLeftClass()

                # list of all glyphs kerned to the left of a pair:
                allLeftGlyphs = pairPos.Coverage.glyphs
                # list of all glyphs contained in left-sided kerning classes:
                # allLeftClassGlyphs = pairPos.ClassDef1.classDefs.keys()

                singleGlyphs = []
                classGlyphs = []

                for gName, classID in pairPos.ClassDef1.classDefs.items():
                    if classID == 0:
                        singleGlyphs.append(gName)
                    else:
                        classGlyphs.append(gName)
                # coverage glyphs minus glyphs in real class (without class 0)
                lg0.glyphs = list(set(allLeftGlyphs) - set(classGlyphs))

                lg0.glyphs.sort()
                leftClasses[lg0.class1Record] = lg0
                className = "class_%s_%s" % (loop, lg0.class1Record)
                self.allLeftClasses[className] = lg0.glyphs

                # Find all the remaining left classes:
                for leftGlyph in pairPos.ClassDef1.classDefs:
                    class1Record = pairPos.ClassDef1.classDefs[leftGlyph]

                    if class1Record != 0:  # this was the crucial line.
                        lg = myLeftClass()
                        lg.class1Record = class1Record
                        leftClasses.setdefault(class1Record, lg).glyphs.append(
                            leftGlyph
                        )
                        self.allLeftClasses.setdefault(
                            "class_%s_%s" % (loop, lg.class1Record), lg.glyphs
                        )

                # Same for the right classes:
                for rightGlyph in pairPos.ClassDef2.classDefs:
                    class2Record = pairPos.ClassDef2.classDefs[rightGlyph]
                    rg = myRightClass()
                    rg.class2Record = class2Record
                    rightClasses.setdefault(class2Record, rg).glyphs.append(
                        rightGlyph
                    )
                    self.allRightClasses.setdefault(
                        "class_%s_%s" % (loop, rg.class2Record), rg.glyphs
                    )

                for record_l in leftClasses:
                    for record_r in rightClasses:
                        if pairPos.Class1Record[record_l].Class2Record[
                            record_r
                        ]:
                            valueFormat = pairPos.ValueFormat1

                            if valueFormat in [4, 5]:
                                kernValue = (
                                    pairPos.Class1Record[record_l]
                                    .Class2Record[record_r]
                                    .Value1.XAdvance
                                )
                            elif valueFormat == 0:
                                # valueFormat zero is caused by a value of <0 0 0 0> on a class-class pair; skip these
                                continue
                            else:
                                continue  # skip the rest

                            if kernValue != 0:
                                leftClassName = "class_%s_%s" % (
                                    loop,
                                    leftClasses[record_l].class1Record,
                                )
                                rightClassName = "class_%s_%s" % (
                                    loop,
                                    rightClasses[record_r].class2Record,
                                )

                                self.classPairs[
                                    (leftClassName, rightClassName)
                                ] = kernValue

                                for l in leftClasses[record_l].glyphs:
                                    for r in rightClasses[record_r].glyphs:
                                        if (l, r) in self.kerningPairs:
                                            # if the kerning pair has already been assigned in pair-to-pair kerning
                                            continue
                                        else:
                                            if valueFormat == 5:  # RTL kerning
                                                kernValue = "<%d 0 %d 0>" % (
                                                    pairPos.Class1Record[
                                                        record_l
                                                    ]
                                                    .Class2Record[record_r]
                                                    .Value1.XPlacement,
                                                    pairPos.Class1Record[
                                                        record_l
                                                    ]
                                                    .Class2Record[record_r]
                                                    .Value1.XAdvance,
                                                )

                                            self.kerningPairs[
                                                (l, r)
                                            ] = kernValue


def for_font(f: ttLib.TTFont) -> Mapping[tuple[str, str], float]:
    reader = OTFKernReader(f)
    glyph_name_to_char = {
        gname: chr(ord_) for ord_, gname in f.getBestCmap().items()
    }
    return {
        (glyph_name_to_char[a], glyph_name_to_char[b]): v
        for (a, b), v in reader.kerningPairs.items()
        if a in glyph_name_to_char and b in glyph_name_to_char
    }



from __future__ import annotations

import abc
from dataclasses import dataclass
from itertools import islice, tee
from typing import Callable, Iterator, Sequence

from ..common import (
    XY,
    Streamable,
    add_slots,
    fix_abstract_properties,
    flatten,
    peek,
    prepend,
)
from ..page import Column, Page
from ..resources import Resources
from ..style import StyleFull
from ..units import Pt

__all__ = [
    "Block",
]


class Block(abc.ABC):
    """Base class for block elements that can be laid out in a column
    by :class:`~pdfje.AutoPage`.
    """

    __slots__ = ()

    # Fill the given columns with this block's content. It may consume as many
    # columns as it needs to determine how to render itself. It should only
    # yield columns that are actually filled -- which may be fewer than it
    # consumed (e.g. if it needed to look ahead).
    #
    # Why not a generator? Because a block may need to consume multiple
    # columns to render itself, before starting to yield completed columns
    @abc.abstractmethod
    def into_columns(
        self, res: Resources, style: StyleFull, cs: Iterator[ColumnFill], /
    ) -> Iterator[ColumnFill]:
        ...


@fix_abstract_properties
class Shaped(abc.ABC):
    __slots__ = ()

    # FUTURE: remove width from this interface. It can be set
    # on this object itself.
    @abc.abstractmethod
    def render(self, pos: XY, width: Pt) -> Streamable:
        ...

    @property
    @abc.abstractmethod
    def height(self) -> Pt:
        ...


@add_slots
@dataclass(frozen=True)
class ColumnFill(Streamable):
    box: Column
    blocks: Sequence[tuple[XY, Shaped]]
    height_free: Pt

    @staticmethod
    def new(col: Column) -> ColumnFill:
        return ColumnFill(col, [], col.height)

    def add(self, s: Shaped) -> ColumnFill:
        return ColumnFill(
            self.box,
            (*self.blocks, (self.cursor(), s)),
            self.height_free - s.height,
        )

    def cursor(self) -> XY:
        return self.box.origin.add_y(self.height_free)

    def __iter__(self) -> Iterator[bytes]:
        for loc, s in self.blocks:
            yield from s.render(loc, self.box.width)


_ColumnFiller = Callable[[Iterator[ColumnFill]], Iterator[ColumnFill]]


@add_slots
@dataclass(frozen=True)
class PageFill:
    base: Page
    todo: Sequence[ColumnFill]  # in the order they will be filled
    done: Sequence[ColumnFill]  # most recently filled last

    def reopen_most_recent_column(self) -> PageFill:
        return PageFill(self.base, (self.done[-1], *self.todo), self.done[:-1])

    @staticmethod
    def new(page: Page) -> PageFill:
        return PageFill(page, list(map(ColumnFill.new, page.columns)), ())


def fill_pages(
    doc: Iterator[PageFill], f: _ColumnFiller
) -> tuple[Iterator[PageFill], Sequence[PageFill]]:
    trunk, branch = tee(doc)
    return _fill_into(  # pragma: no branch
        f(flatten(p.todo for p in branch)), trunk
    )


def _fill_into(
    filled: Iterator[ColumnFill], doc: Iterator[PageFill]
) -> tuple[Iterator[PageFill], Sequence[PageFill]]:
    try:
        _, filled = peek(filled)
    except StopIteration:
        return doc, []  # no content to add

    completed: list[PageFill] = []
    for page in doc:  # pragma: no branch
        page_cols = list(islice(filled, len(page.todo)))
        completed.append(
            PageFill(
                page.base,
                page.todo[len(page_cols) :],  # noqa
                (*page.done, *page_cols),
            )
        )
        try:
            _, filled = peek(filled)
        except StopIteration:
            break  # no more content -- wrap things up

    return prepend(completed.pop().reopen_most_recent_column(), doc), completed
